#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Paging list related class."""

from bisect import bisect_right
from functools import partial
from itertools import accumulate, chain, repeat
from math import ceil
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

from graviti.utility.typing import NestedDict

_P = TypeVar("_P", bound="PagingList")
_O = TypeVar("_O", bound="Offsets")


class Offsets:
    """The offsets manager of the paging list.

    Arguments:
        total_count: The total count of the elements in the paging list.
        limit: The size of each page.

    """

    _offsets: List[int]

    def __init__(self, total_count: int, limit: int) -> None:
        self.total_count = total_count
        self._limit = limit

    def _get_offsets(self) -> List[int]:
        if not hasattr(self, "_offsets"):
            self._offsets = (
                list(range(0, self.total_count, self._limit)) if self._limit != 0 else []
            )

        return self._offsets

    def update(self, start: int, stop: int, lengths: Iterable[int]) -> None:
        """Update the offsets when setting or deleting paging list items.

        Arguments:
            start: The start index.
            stop: The stop index.
            lengths: The length of the set values.

        """
        offsets = self._get_offsets()
        partial_offsets = list(accumulate(chain((offsets[start],), lengths)))
        try:
            last_offset = offsets[stop + 1]
        except IndexError:
            last_offset = self.total_count

        diff = partial_offsets.pop() - last_offset
        if diff != 0:
            self.total_count += diff
            for i in range(start + 1, len(offsets)):
                offsets[i] += diff

        offsets[start : stop + 1] = partial_offsets

    def get_coordinate(self, index: int) -> Tuple[int, int]:
        """Get the page coordinate of the elements.

        Arguments:
            index: The index of the element in paging list.

        Returns:
            The page number and the index of the page.

        """
        if not hasattr(self, "_offsets"):
            return divmod(index, self._limit)

        i = bisect_right(self._offsets, index) - 1
        return i, index - self._offsets[i]

    def extend(self, lengths: Iterable[int]) -> None:
        """Update the offsets when extending the paging list.

        Arguments:
            lengths: The lengths of the extended pages.

        """
        offsets = self._get_offsets()
        offsets.extend(accumulate(chain((self.total_count,), lengths)))
        self.total_count = offsets.pop()

    def copy(self: _O) -> _O:
        """Return a copy of the Offsets.

        Returns:
            A copy of the Offsets.

        """
        obj = self.__class__(self.total_count, self._limit)
        if hasattr(self, "_offsets"):
            obj._offsets = self._offsets.copy()  # pylint: disable=protected-access

        return obj


class Page:
    """Page is an array wrapper and represents a page in paging list.

    Arguments:
        array: The pyarrow array.

    """

    def __init__(self, array: pa.Array):
        self._ranging = range(len(array))
        self._patch(array)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._ranging})"

    def __len__(self) -> int:
        return len(self._ranging)

    def __iter__(self) -> Iterator[Any]:
        return self._iter()

    @overload
    def __getitem__(self, index: int) -> Any:
        ...

    @overload
    def __getitem__(self, index: slice) -> "Page":
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[Any, "Page"]:
        if isinstance(index, slice):
            return self.get_slice(index.start, index.stop, index.step)

        return self.get_item(index)

    def _patch(self, array: pa.Array) -> None:
        self._array = array
        # https://github.com/python/mypy/issues/708
        # https://github.com/python/mypy/issues/2427
        self._iter = array.__iter__  # type: ignore[assignment]
        self.get_item = array.__getitem__  # type: ignore[assignment]

    def _iter(self) -> Iterator[Any]:  # pylint: disable=method-hidden
        return self.get_array().__iter__()  # type: ignore[no-any-return]

    def get_item(self, index: int) -> Any:  # pylint: disable=method-hidden
        """Return the item at the given index.

        Arguments:
            index: Position of the mutable sequence.

        Returns:
            The item at the given index.

        """
        return self.get_array().__getitem__(index)

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "Page":
        """Return a sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return SlicedPage(self._ranging[start:stop:step], self._array)

    def get_array(self) -> pa.array:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        return self._array


class SlicedPage(Page):
    """SlicedPage is an array wrapper and represents a sliced page in paging list.

    Arguments:
        ranging: The range instance of this page.
        array: The pyarrow array.

    """

    def __init__(  # pylint: disable=super-init-not-called
        self, ranging: range, array: pa.Array
    ) -> None:
        self._ranging = ranging
        self._array = None
        self._source_array = array

    def get_array(self) -> pa.array:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        if self._array is None:
            ranging = self._ranging
            array = self._source_array[ranging.start : ranging.stop : ranging.step]
            self._patch(array)

        return self._array


class LazyPage(Page):
    """LazyPage is a placeholder when the paging list page is not loaded yet.

    Arguments:
        pos: The page number.
        ranging: The range instance of this page.
        parent: The parent paging list.

    """

    def __init__(  # pylint: disable=super-init-not-called
        self, ranging: range, pos: int, array_getter: Callable[[int], pa.Array]
    ) -> None:
        self._ranging = ranging
        self._pos = pos
        self._array_getter = array_getter
        self._array = None

    def get_slice(
        self, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None
    ) -> "LazySlicedPage":
        """Return a lazy sliced page according to the given start and stop index.

        Arguments:
            start: The start index.
            stop: The stop index.
            step: The slice step.

        Returns:
            A sliced page according to the given start and stop index.

        """
        return LazySlicedPage(self._ranging[start:stop:step], self._pos, self._array_getter)

    def get_array(self) -> pa.Array:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        if self._array is None:
            array = self._array_getter(self._pos)
            self._patch(array)

        return self._array


class LazySlicedPage(LazyPage):
    """LazySlicedPage is a placeholder when the sliced paging list page is not loaded yet."""

    def get_array(self) -> pa.Array:
        """Get the array inside the page.

        Returns:
            The array inside the page.

        """
        if self._array is None:
            ranging = self._ranging
            array = self._array_getter(self._pos)[ranging.start : ranging.stop : ranging.step]

            self._patch(array)

        return self._array


class PagingList:
    """PagingList is a list composed of multiple lists (pages).

    Arguments:
        array: The input pyarrow array.

    """

    def __init__(self, array: pa.Array) -> None:
        length = len(array)
        self._pages = [Page(array)] if length != 0 else []
        self._offsets = Offsets(length, length)
        self._patype = array.type

    def __len__(self) -> int:
        return self._offsets.total_count

    def __iter__(self) -> Iterator[Any]:
        return chain(*self._pages)

    @overload
    def __setitem__(self, index: int, value: Any) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Union[Iterable[Any], "PagingList"]) -> None:
        ...

    def __setitem__(
        self, index: Union[int, slice], value: Union[Any, Iterable[Any], "PagingList"]
    ) -> None:
        if isinstance(index, int):
            self.set_item(index, value)
        elif isinstance(value, PagingList):
            self.set_slice(index, value)
        else:
            self.set_slice_iterable(index, value)

    def __delitem__(self, index: Union[int, slice]) -> None:
        if isinstance(index, slice):
            start, stop, step = index.indices(self.__len__())

            if step == 1:
                pass
            elif step == -1:
                start, stop = stop + 1, start + 1
            else:
                ranging: Any = range(start, stop, step)
                if step > 0:
                    ranging = reversed(ranging)

                for i in ranging:
                    self._update_pages(i, i + 1)

                return

        else:
            index = self._make_index_nonnegative(index)
            start = index
            stop = index + 1

        self._update_pages(start, stop)

    def __getitem__(self, index: int) -> Any:
        index = self._make_index_nonnegative(index)
        return self._getitem(index)

    def __iadd__(self: _P, values: Union["PagingList", Iterable[Any]]) -> _P:
        if isinstance(values, self.__class__):
            self.extend(values)
        else:
            self.extend_iterable(values)

        return self

    def _make_index_nonnegative(self, index: int) -> int:
        return index if index >= 0 else self.__len__() + index

    def _getitem(self, index: int) -> Any:
        i, j = self._offsets.get_coordinate(index)
        return self._pages[i].get_item(j)

    def _update_pages(self, start: int, stop: int, pages: Optional[Sequence[Page]] = None) -> None:
        if start >= stop and not pages:
            return

        stop = max(start, stop)

        start_i, start_j = self._offsets.get_coordinate(start)
        stop_i, stop_j = self._offsets.get_coordinate(stop - 1)

        update_pages = []
        update_lengths = []

        left_page = self._pages[start_i].get_slice(stop=start_j)
        left_length = len(left_page)
        if left_length:
            update_pages.append(left_page)
            update_lengths.append(left_length)

        if pages:
            update_pages.extend(pages)
            update_lengths.extend(map(len, pages))

        right_page = self._pages[stop_i].get_slice(stop_j + 1)
        right_length = len(right_page)
        if right_length:
            update_pages.append(right_page)
            update_lengths.append(right_length)

        self._pages[start_i : stop_i + 1] = update_pages
        self._offsets.update(start_i, stop_i, update_lengths)

    def _update_pages_with_step(
        self, start: int, stop: int, step: int, values: "PagingList"
    ) -> None:
        length = len(values)
        ranging: Any = range(start, stop, step)
        indices: Any = range(length)

        if length != len(ranging):
            raise ValueError(
                f"attempt to assign sequence of size {length} "
                f"to extended slice of size {len(ranging)}"
            )

        if step > 0:
            ranging = reversed(ranging)
            indices = reversed(indices)

        offsets = values._offsets  # pylint: disable=protected-access
        pages = values._pages  # pylint: disable=protected-access

        for i, j in zip(ranging, indices):
            x, y = offsets.get_coordinate(j)
            self._update_pages(i, i + 1, [pages[x][y : y + 1]])

    @classmethod
    def from_factory(
        cls: Type[_P], factory: "LazyFactory", keys: Tuple[str, ...], patype: pa.DataType
    ) -> _P:
        """Create PagingList from LazyFactory.

        Arguments:
            factory: The parent :class:`LazyFactory` instance.
            keys: The keys to access the array from factory.
            patype: The pyarrow DataType of the elements in the list.

        Returns:
            The PagingList instance created from given factory.

        """
        obj: _P = object.__new__(cls)
        array_getter = partial(factory.get_array, keys=keys)
        obj._pages = [
            LazyPage(ranging, pos, array_getter)
            for pos, ranging in enumerate(factory.get_page_ranges())
        ]
        obj._offsets = Offsets(
            factory._total_count, factory._limit  # pylint: disable=protected-access
        )
        obj._patype = patype

        return obj

    def set_item(self, index: int, value: Any) -> None:
        """Update the element value in PagingList at the given index.

        Arguments:
            index: The element index.
            value: The value needs to be set into the PagingList.

        """
        index = self._make_index_nonnegative(index)
        page = Page(pa.array((value,), self._patype))
        self._update_pages(index, index + 1, [page])

    def set_slice(self, index: slice, values: "PagingList") -> None:
        """Update the element values in PagingList at the given slice with another PagingList.

        Arguments:
            index: The element slice.
            values: The PagingList which contains the elements to be set.

        Raises:
            ArrowTypeError: When two pyarrow types mismatch.
            ValueError: When the input size mismatches with the slice size (when step != 1).

        """
        # pylint: disable=protected-access
        if values._patype != self._patype:
            raise pa.ArrowTypeError(
                f"Can not set a '{self._patype}' list with a '{values._patype}' list"
            )

        start, stop, step = index.indices(self.__len__())

        if step == 1:
            self._update_pages(start, stop, values._pages)
            return

        if step == -1:
            start, stop = stop + 1, max(start, stop) + 1
            if len(values) != stop - start:
                raise ValueError(
                    f"attempt to assign sequence of size {len(values)} "
                    f"to extended slice of size {stop - start}"
                )

            self._update_pages(
                start, stop, [page.get_slice(step=-1) for page in reversed(values._pages)]
            )
            return

        self._update_pages_with_step(start, stop, step, values)

    def set_slice_iterable(self, index: slice, values: Iterable[Any]) -> None:
        """Update the element values in PagingList at the given slice with iterable object.

        Arguments:
            index: The element slice.
            values: The iterable object which contains the elements to be set.

        Raises:
            ValueError: When the assign input size mismatches with the slice size (when step != 1).

        """
        start, stop, step = index.indices(self.__len__())

        if step == 1:
            array = pa.array(values, self._patype)
            self._update_pages(start, stop, [Page(array)] if len(array) != 0 else None)
            return

        if step == -1:
            try:
                values = reversed(values)  # type: ignore[call-overload]
            except TypeError:
                values = reversed(list(values))

            array = pa.array(values, self._patype)

            start, stop = stop + 1, max(start, stop) + 1
            if len(array) != stop - start:
                raise ValueError(
                    f"attempt to assign sequence of size {len(array)} "
                    f"to extended slice of size {stop - start}"
                )

            self._update_pages(start, stop, [Page(array)] if len(array) != 0 else None)
            return

        self._update_pages_with_step(start, stop, step, PagingList(pa.array(values, self._patype)))

    def extend(self, values: "PagingList") -> None:
        """Extend PagingList by appending elements from another PagingList.

        Arguments:
            values: The PagingList which contains the elements to be extended.

        Raises:
            ArrowTypeError: When two pyarrow types mismatch.

        """
        # pylint: disable=protected-access
        if values._patype != self._patype:
            raise pa.ArrowTypeError(
                f"Can not extend a '{self._patype}' list with a '{values._patype}' list"
            )

        pages = values._pages
        self._offsets.extend(map(len, pages))
        self._pages.extend(pages)

    def extend_iterable(self, values: Iterable[Any]) -> None:
        """Extend PagingList by appending elements from the iterable.

        Arguments:
            values: Elements to be extended into the PagingList.

        """
        page = Page(pa.array(values, self._patype))
        self._offsets.extend((len(page),))
        self._pages.append(page)

    def copy(self: _P) -> _P:
        """Return a copy of the paging list.

        Returns:
            A copy of the paging list.

        """
        obj: _P = object.__new__(self.__class__)
        # pylint: disable=protected-access
        obj._pages = self._pages.copy()
        obj._offsets = self._offsets.copy()
        obj._patype = self._patype
        return obj

    def to_pyarrow(self) -> pa.ChunkedArray:
        """Convert the paging list to pyarrow ChunkedArray.

        Returns:
            The pyarrow ChunkedArray.

        """
        return pa.chunked_array((page.get_array() for page in self._pages), self._patype)


PagingLists = NestedDict[str, PagingList]


class LazyFactory:
    """LazyFactory is a factory for creating paging lists.

    Arguments:
        total_count: The total count of the elements in the paging lists.
        limit: The size of each lazy load page.
        getter: A callable object to get the source data.
        patype: The pyarrow DataType of the data in the factory.

    Examples:
        >>> import pyarrow as pa
        >>> patype = pa.struct(
        ...     {
        ...         "remotePath": pa.string(),
        ...         "label": pa.struct({"CLASSIFICATION": pa.struct({"category": pa.string()})}),
        ...     }
        ... )
        >>> TOTAL_COUNT = 1000
        >>> def getter(offset: int, limit: int) -> List[Dict[str, Any]]:
        ...     stop = min(offset + limit, TOTAL_COUNT)
        ...     return [
        ...         {
        ...             "remotePath": f"{i:06}.jpg",
        ...             "label": {"CLASSIFICATION": {"category": "cat" if i % 2 else "dog"}},
        ...         }
        ...         for i in range(offset, stop)
        ...     ]
        ...
        >>> factory = LazyFactory(TOTAL_COUNT, 128, getter, patype)
        >>> paths = factory.create_list(("remotePath",))
        >>> categories = factory.create_list(("label", "CLASSIFICATION", "category"))
        >>> len(paths)
        1000
        >>> list(paths)
        [<pyarrow.StringScalar: '000000.jpg'>,
         <pyarrow.StringScalar: '000001.jpg'>,
         <pyarrow.StringScalar: '000002.jpg'>,
         <pyarrow.StringScalar: '000003.jpg'>,
         <pyarrow.StringScalar: '000004.jpg'>,
         <pyarrow.StringScalar: '000005.jpg'>,
         ...
         ...
         <pyarrow.StringScalar: '000999.jpg'>]
        >>> len(categories)
        1000
        >>> list(categories)
        [<pyarrow.StringScalar: 'dog'>,
         <pyarrow.StringScalar: 'cat'>,
         <pyarrow.StringScalar: 'dog'>,
         <pyarrow.StringScalar: 'cat'>,
         <pyarrow.StringScalar: 'dog'>,
         ...
         ...
         <pyarrow.StringScalar: 'cat'>]


    """

    def __init__(
        self, total_count: int, limit: int, getter: Callable[[int, int], Any], patype: pa.DataType
    ) -> None:
        self._getter = getter
        self._total_count = total_count
        self._limit = limit
        self._patype = patype
        self._pages: List[Optional[pa.StructArray]] = [None] * ceil(total_count / limit)

    def get_array(self, pos: int, keys: Tuple[str, ...]) -> pa.Array:
        """Get the array from the factory.

        Arguments:
            pos: The page number.
            keys: The keys to access the array from factory.

        Returns:
            The requested pyarrow array.

        """
        array = self._pages[pos]
        if array is None:
            array = pa.array(self._getter(pos * self._limit, self._limit), type=self._patype)
            self._pages[pos] = array

        for key in keys:
            array = array.field(key)

        return array

    def create_list(self, keys: Tuple[str, ...]) -> PagingList:
        """Create a paging list from the factory.

        Arguments:
            keys: The keys to access the array from factory.

        Returns:
            A paging list created by the given keys.

        """
        patype = self._patype
        for key in keys:
            patype = patype[key].type

        return PagingList.from_factory(self, keys, patype)

    def create_lists(self, keys: List[Tuple[str, ...]]) -> PagingLists:
        """Create a dict of PagingList from the given keys.

        Arguments:
            keys: A list of keys to create the paging lists.

        Returns:
            The created paging lists.

        """
        paging_lists: PagingLists = {}
        for key in keys:
            position = paging_lists
            for subkey in key[:-1]:
                position = position.setdefault(subkey, {})  # type: ignore[assignment]
            position[key[-1]] = self.create_list(key)

        return paging_lists

    def get_page_ranges(self) -> Iterator[range]:
        """A Generator which generates the range of the pages in the factory.

        Yields:
            The page ranges.

        """
        div, mod = divmod(self._total_count, self._limit)
        yield from repeat(range(self._limit), div)
        if mod != 0:
            yield range(mod)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Schema to colomn extractors related methods."""


from functools import partial
from typing import Any, Callable, Dict, Iterator, List, Tuple, TypeVar, Union

import pyarrow as pa

from graviti.utility import NestedDict, RemoteFileType

_T = TypeVar("_T")
_D = Dict[str, Tuple[Iterator[_T], str]]
_Extractor = Tuple[Callable[[Dict[str, Any]], Iterator[Any]], _T]
Extractors = NestedDict[str, _Extractor[Any]]

_PORTEX_TYPE_TO_PYARROW_DTYPE = {
    "boolean": pa.bool_(),
    "binary": pa.binary(),
    "string": pa.string(),
    "enum": pa.string(),
}


def _get_filename(*_: Any) -> _Extractor[str]:
    def extractor(data: Dict[str, Any]) -> Iterator[str]:
        for item in data["dataDetails"]:
            yield item["remotePath"]

    return extractor, pa.string()


def _get_file(*_: Any) -> _Extractor[str]:
    def extractor(data: Dict[str, Any]) -> Iterator[Tuple[str, str]]:
        for item in data["dataDetails"]:
            yield item["url"], item["checksum"]

    return extractor, RemoteFileType()


def _get_category(*_: Any) -> _Extractor[Any]:
    def extractor(data: Dict[str, Any]) -> Iterator[Any]:
        for item in data["dataDetails"]:
            yield item["label"]["CLASSIFICATION"]["category"]

    return extractor, pa.string()


def _attribute_extractor(data: Dict[str, Any], name: str) -> Iterator[Any]:
    for item in data["dataDetails"]:
        yield item["label"]["CLASSIFICATION"]["attributes"][name]


def _get_attribute(schema: Dict[str, Any]) -> Extractors:
    attributes: Extractors = {}
    for attribute_info in schema["attributes"]:
        name = attribute_info["name"]
        type_ = attribute_info["type"]
        dtype = _PORTEX_TYPE_TO_PYARROW_DTYPE.get(type_, type_)
        attributes[name] = partial(_attribute_extractor, name=name), dtype
    return attributes


def _extract_attribute(data: List[Dict[str, Any]], schema: Dict[str, Any]) -> Dict[str, List[Any]]:
    attributes = {}
    for attribute_info in schema["attributes"]:
        name = attribute_info["name"]
        attributes[name] = [item["attributes"][name] for item in data]
    return attributes


def _extract_category_and_attribute(
    schema: Dict[str, Any], labels: List[Dict[str, Any]]
) -> Dict[str, Any]:
    results: Dict[str, Any] = {}
    if "categories" in schema:
        results["category"] = [label["category"] for label in labels]
    if "attributes" in schema:
        results["attribute"] = _extract_attribute(labels, schema)
    return results


def _get_box2ds(schema: Dict[str, Any]) -> _Extractor[Dict[str, Any]]:
    def extractor(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for item in data["dataDetails"]:
            labels = item["label"]["BOX2D"]
            box2d: Dict[str, Any] = {
                "xmin": [label["box2d"]["xmin"] for label in labels],
                "xmax": [label["box2d"]["xmax"] for label in labels],
                "ymin": [label["box2d"]["ymin"] for label in labels],
                "ymax": [label["box2d"]["ymax"] for label in labels],
            }
            box2d.update(_extract_category_and_attribute(schema["items"], labels))
            yield box2d

    return extractor, pa.string()


def _extract_vetices(
    pointlist: List[Dict[str, Any]], vertices_keys: Tuple[str, ...]
) -> Dict[str, List[Any]]:
    return {key: [point[key] for point in pointlist] for key in vertices_keys}


def _get_pointlists(schema: Dict[str, Any], key: str) -> _Extractor[Dict[str, Any]]:
    def extractor(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for item in data["dataDetails"]:
            labels = item["label"][key.upper()]
            pointlists: Dict[str, Any] = {
                "vertices": [_extract_vetices(label[key], ("x", "y")) for label in labels]
            }
            pointlists.update(_extract_category_and_attribute(schema["items"], labels))
            yield pointlists

    return extractor, pa.string()


def _get_keypoints2ds(schema: Dict[str, Any]) -> _Extractor[Dict[str, Any]]:
    keypoints2d_schema = schema["items"]
    vertices_keys = ("x", "y", "v") if keypoints2d_schema.get("visible") else ("x", "y")
    vertices_extractor = partial(_extract_vetices, vertices_keys=vertices_keys)

    def extractor(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for item in data["dataDetails"]:
            labels = item["label"]["KEYPOINTS2D"]
            keypoints2d: Dict[str, Any] = {
                "vertices": [vertices_extractor(label["keypoints2d"]) for label in labels]
            }
            keypoints2d.update(_extract_category_and_attribute(keypoints2d_schema, labels))
            yield keypoints2d

    return extractor, pa.string()


def _mask_extractor(data: Dict[str, Any], key: str) -> Iterator[Any]:
    for item in data["dataDetails"]:
        label = item["label"][key]
        yield label["url"]


def _info_extractor(
    data: Dict[str, Any], schema: Dict[str, Any], key: str
) -> Iterator[Dict[str, Any]]:
    id_key = "categoryId" if key == "SEMANTIC_MASK" else "instanceId"
    for item in data["dataDetails"]:
        all_info = item["label"][key]["info"]
        mask_info = {
            "id": [info[id_key] for info in all_info],
            "attribute": _extract_attribute(all_info, schema),
        }
        if key == "PANOPTIC_MASK":
            mask_info["categoryId"] = [info["categoryId"] for info in all_info]
        yield mask_info


def _get_mask(schema: Dict[str, Any], key: str) -> Extractors:
    mask: Dict[str, Any] = {"mask": (partial(_mask_extractor, key=key), "object")}
    if "attributes" in schema:
        mask["info"] = partial(_info_extractor, schema=schema, key=key), "object"
    return mask


def _get_multi_pointlists(schema: Dict[str, Any], key: str) -> _Extractor[Dict[str, Any]]:
    def extractor(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for item in data["dataDetails"]:
            labels = item["label"][f"MULTI_{key.upper()}"]
            multi_pointlists: Dict[str, Any] = {
                f"{key}s": [
                    [
                        _extract_vetices(pointlist, ("x", "y"))
                        for pointlist in label[f"multi{key.capitalize()}"]
                    ]
                    for label in labels
                ]
            }
            multi_pointlists.update(_extract_category_and_attribute(schema["items"], labels))
            yield multi_pointlists

    return extractor, pa.string()


def _get_rle(schema: Dict[str, Any]) -> _Extractor[Dict[str, Any]]:
    def extractor(data: Dict[str, Any]) -> Iterator[Dict[str, Any]]:
        for item in data["dataDetails"]:
            labels = item["label"]["RLE"]
            rle: Dict[str, Any] = {"code": [label["rle"] for label in labels]}
            rle.update(_extract_category_and_attribute(schema["items"], labels))
            yield rle

    return extractor, pa.string()


_EXTRACTORS_GETTER: Dict[
    Tuple[str, str], Callable[[Dict[str, Any]], Union[_Extractor[Any], Extractors]]
] = {
    ("filename", "string"): _get_filename,
    ("image", "file.Image"): _get_file,
    ("point_cloud", "file.PointCloud"): _get_file,
    ("point_cloud", "file.PointCloudBin"): _get_file,
    ("audio", "file.Audio"): _get_file,
    ("file", "binary"): _get_file,
    ("attribute", "label.Attribute"): _get_attribute,
    ("category", "label.Category"): _get_category,
    ("box2ds", "array"): _get_box2ds,
    ("polygons", "array"): partial(_get_pointlists, key="polygon"),
    ("polyline2ds", "array"): partial(_get_pointlists, key="polyline2d"),
    ("keypoints2ds", "array"): _get_keypoints2ds,
    ("semantic_mask", "label.file.SemanticMask"): partial(_get_mask, key="SEMANTIC_MASK"),
    ("instance_mask", "label.file.InstanceMask"): partial(_get_mask, key="INSTANCE_MASK"),
    ("panoptic_mask", "label.file.PanopticMask"): partial(_get_mask, key="PANOPTIC_MASK"),
    ("multi_polygons", "array"): partial(_get_multi_pointlists, key="polygon"),
    ("multi_polyline2ds", "array"): partial(_get_multi_pointlists, key="polyline2d"),
    ("RLE", "array"): _get_rle,
}


def get_extractors(schema: Dict[str, Any]) -> Extractors:
    """Get the extractors and dtypes for colomns.

    Arguments:
        schema: The schema of a DataFrame.

    Returns:
        A dict containing the extractors and dtypes for all colomns.

    Examples:
        >>> import yaml
        >>>
        >>> from graviti.client import list_data_details
        >>> from graviti.utility.lazy import LazyFactory, LazyList
        >>> from graviti.portex import catalog_to_schema, get_extractors
        >>>
        >>> from tensorbay import GAS
        >>> from tensorbay.dataset import Dataset
        >>> ACCESSKEY = "ACCESSKEY-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        >>> URL = "https://gas.graviti.com/"
        >>> DATASET_NAME = "MNIST"
        >>> TOTAL_COUNT = 1000
        >>>
        >>> gas = GAS(ACCESSKEY)
        >>> dataset = Dataset(DATASET_NAME, gas)
        >>> dataset_client = gas.get_dataset(DATASET_NAME)
        >>>
        >>> getter = lambda offset, limit: list_data_details(
        ...     url=URL,
        ...     access_key=ACCESSKEY,
        ...     dataset_id=dataset_client.dataset_id,
        ...     segment_name="train",
        ...     commit=dataset_client.status.commit_id,
        ...     offset=offset,
        ...     limit=limit,
        ... )
        >>> factory = LazyFactory(TOTAL_COUNT, 128, getter)
        >>> schema = yaml.load(
        ...    catalog_to_schema(dataset.catalog, dataset["train"][0], dataset.notes), yaml.Loader
        ... )
        >>> extractors = get_extractors(schema)
        >>> lazy_lists = {}
        >>> for key, arguments in extractors.items():
        ...     lazy_lists[key] = factory.create_list(*arguments)
        >>> lazy_lists
        {'filename': LazyList [
           'train_image_00000.png',
           'train_image_00001.png',
           'train_image_00002.png',
           'train_image_00003.png',
           'train_image_00004.png',
           'train_image_00005.png',
           'train_image_00006.png',
           'train_image_00007.png',
           'train_image_00008.png',
           'train_image_00009.png',
           'train_image_00010.png',
           'train_image_00011.png',
           'train_image_00012.png',
           'train_image_00013.png',
           ... (985 items are folded),
           'train_image_00999.png'
         ],
         'image': LazyList [
           RemoteFileMixin("train_image_00000.png"),
           RemoteFileMixin("train_image_00001.png"),
           RemoteFileMixin("train_image_00002.png"),
           RemoteFileMixin("train_image_00003.png"),
           RemoteFileMixin("train_image_00004.png"),
           RemoteFileMixin("train_image_00005.png"),
           RemoteFileMixin("train_image_00006.png"),
           RemoteFileMixin("train_image_00007.png"),
           RemoteFileMixin("train_image_00008.png"),
           RemoteFileMixin("train_image_00009.png"),
           RemoteFileMixin("train_image_00010.png"),
           RemoteFileMixin("train_image_00011.png"),
           RemoteFileMixin("train_image_00012.png"),
           RemoteFileMixin("train_image_00013.png"),
           ... (985 items are folded),
           RemoteFileMixin("train_image_00999.png")
         ],
         'category': LazyList [
           '5',
           '0',
           '4',
           '1',
           '9',
           '2',
           '1',
           '3',
           '1',
           '4',
           '3',
           '5',
           '3',
           '6',
           ... (985 items are folded),
           '6'
         ]}

    """
    extractors: Extractors = {}
    for field in schema["fields"]:
        field_name, field_type = field["name"], field["type"]
        extractor_getter = _EXTRACTORS_GETTER[field_name, field_type]
        extractors[field_name] = extractor_getter(field)
    return extractors



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Definitions of different operations on a DataFrame."""

from typing import TYPE_CHECKING, List

from graviti.openapi import add_data, update_data, upload_files
from graviti.utility import chunked

if TYPE_CHECKING:
    from graviti.dataframe import Container, DataFrame

_MAX_BATCH_SIZE = 2048


class DataFrameOperation:
    """This class defines the basic method of the operation on a DataFrame."""

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
    ) -> None:
        """Execute the OpenAPI create sheet.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError


class AddData(DataFrameOperation):
    """This class defines the operation that add data to a DataFrame.

    Arguments:
        data: The data to be added.

    """

    def __init__(self, data: "DataFrame") -> None:
        self.data = data

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.

        """
        dataframe = self.data
        arrays = _get_arrays(dataframe, "file.RemoteFile")
        data = dataframe.to_pylist()

        for batch, *file_arrays in zip(
            *map(lambda x: chunked(x, _MAX_BATCH_SIZE), (data, *arrays))  # type: ignore[arg-type]
        ):
            for file_array in file_arrays:
                upload_files(
                    access_key,
                    url,
                    owner,
                    dataset,
                    draft_number=draft_number,
                    sheet=sheet,
                    files=file_array,
                    jobs=jobs,
                )

            add_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch,
            )


class UpdateData(DataFrameOperation):
    """This class defines the operation that updates the data of a DataFrame.

    Arguments:
        data: The data for updating.

    """

    def __init__(self, data: "DataFrame") -> None:
        self.data = data

    def do(  # pylint: disable=invalid-name
        self,
        access_key: str,
        url: str,
        owner: str,
        dataset: str,
        *,
        draft_number: int,
        sheet: str,
        jobs: int,
    ) -> None:
        """Execute the OpenAPI add data.

        Arguments:
            access_key: User's access key.
            url: The URL of the graviti website.
            owner: The owner of the dataset.
            dataset: Name of the dataset, unique for a user.
            draft_number: The draft number.
            sheet: The sheet name.
            jobs: The number of the max workers in multi-thread operation.

        """
        dataframe = self.data
        arrays = _get_arrays(dataframe, "file.RemoteFile")
        data = dataframe.to_pylist()

        for batch, *file_arrays in zip(
            *map(lambda x: chunked(x, _MAX_BATCH_SIZE), (data, *arrays))  # type: ignore[arg-type]
        ):
            for file_array in file_arrays:
                upload_files(
                    access_key,
                    url,
                    owner,
                    dataset,
                    draft_number=draft_number,
                    sheet=sheet,
                    files=file_array,
                    jobs=jobs,
                )

            update_data(
                access_key,
                url,
                owner,
                dataset,
                draft_number=draft_number,
                sheet=sheet,
                data=batch,
            )


def _get_arrays(dataframe: "DataFrame", type_name: str) -> List["Container"]:
    arrays = []
    for key in dataframe.schema.get_keys(type_name):
        array: "Container" = dataframe
        for subkey in key:
            array = array[subkey]  # type: ignore[index]
        arrays.append(array)
    return arrays



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Utility module."""

from graviti.utility.attr import AttrDict
from graviti.utility.collections import (
    NameOrderedDict,
    UserMapping,
    UserMutableMapping,
    UserMutableSequence,
    UserSequence,
)
from graviti.utility.common import locked, shorten, urlnorm
from graviti.utility.file import File, FileBase, RemoteFile
from graviti.utility.itertools import chunked
from graviti.utility.paging import LazyFactory, PagingList, PagingLists
from graviti.utility.pyarrow import (
    BuiltinExtension,
    ExtensionBase,
    ExternalExtension,
    GravitiExtension,
    RemoteFileArray,
    RemoteFileType,
)
from graviti.utility.repr import MAX_REPR_ROWS, ReprMixin, ReprType
from graviti.utility.requests import config, get_session, submit_multithread_tasks
from graviti.utility.typing import NestedDict, PathLike

__all__ = [
    "AttrDict",
    "BuiltinExtension",
    "ExtensionBase",
    "ExternalExtension",
    "File",
    "FileBase",
    "GravitiExtension",
    "LazyFactory",
    "MAX_REPR_ROWS",
    "NameOrderedDict",
    "NestedDict",
    "PagingList",
    "PagingLists",
    "PathLike",
    "RemoteFile",
    "RemoteFileArray",
    "RemoteFileType",
    "ReprMixin",
    "ReprType",
    "UserMapping",
    "UserMutableMapping",
    "UserMutableSequence",
    "UserSequence",
    "chunked",
    "config",
    "get_session",
    "locked",
    "urlnorm",
    "shorten",
    "submit_multithread_tasks",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti indexing related class."""

from typing import TYPE_CHECKING, Any, Iterable, TypeVar, Union, overload

from graviti.utility import NestedDict

if TYPE_CHECKING:
    from graviti.dataframe import DataFrame
    from graviti.dataframe.row.series import Series as RowSeries

_T = TypeVar("_T", int, str)


class DataFrameILocIndexer:
    """Index class for DataFrame.iloc."""

    def __init__(self, obj: "DataFrame") -> None:
        self.obj = obj

    # @overload
    # def __getitem__(self, key: Union[Tuple[int, str], Iterable[bool], slice]) -> Any:
    #     ...

    # @overload
    # def __getitem__(self, key: int) -> "RowSeries":
    #     ...

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "DataFrame":
    #     ...

    def __getitem__(self, key: int) -> "RowSeries":
        return self.obj._getitem_by_location(key)

    @overload
    def __setitem__(self, key: int, value: NestedDict[str, Any]) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[NestedDict[str, Any]]) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[NestedDict[str, Any], Iterable[NestedDict[str, Any]]],
    ) -> None:
        pass


class DataFrameLocIndexer:
    """Index class for DataFrame.loc."""

    def __init__(self, obj: "DataFrame") -> None:
        self.obj: "DataFrame" = obj

    # @overload
    # def __getitem__(self, key: Union[Tuple[int, str], Iterable[bool], slice]) -> Any:
    #     ...

    # @overload
    # def __getitem__(self, key: int) -> "RowSeries":
    #     ...

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "DataFrame":
    #     ...

    def __getitem__(self, key: int) -> "RowSeries":
        return self.obj._getitem_by_location(key)

    @overload
    def __setitem__(self, key: _T, value: NestedDict[str, Any]) -> None:
        ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[NestedDict[str, Any]]) -> None:
        ...

    def __setitem__(
        self,
        key: Union[_T, slice],
        value: Union[NestedDict[str, Any], Iterable[NestedDict[str, Any]]],
    ) -> None:
        pass



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Branch and BranchManager."""

from functools import partial
from typing import TYPE_CHECKING, Generator, Optional

from tensorbay.utility import attr

from graviti.exception import ResourceNotExistError
from graviti.manager.commit import NamedCommit
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import create_branch, delete_branch, get_branch, list_branches

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset

_ERROR_MESSAGE = "The '{attr_name}' is not available due to this branch has no commit history."
_attr = partial(attr, is_dynamic=True, error_message=_ERROR_MESSAGE)


class Branch(NamedCommit):
    """This class defines the structure of a branch.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the branch.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.
        committed_at: The time when the draft is committed.

    """

    parent_commit_id: str = _attr()
    title: str = _attr()
    description: str = _attr()
    committer: str = _attr()
    committed_at: str = _attr()


class BranchManager:
    """This class defines the operations on the branch in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(self, offset: int, limit: int) -> Generator[Branch, None, int]:
        response = list_branches(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            offset=offset,
            limit=limit,
        )

        for item in response["branches"]:
            yield Branch(self._dataset, **item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, name: str, revision: Optional[str] = None) -> Branch:
        """Create a branch.

        Arguments:
            name: The branch name.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If the revision is not given, create the branch based on the current commit.

        Returns:
            The :class:`.Branch` instance with the given name.

        """
        if not revision:
            revision = self._dataset.HEAD.commit_id

        response = create_branch(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            name=name,
            revision=revision,
        )
        return Branch(self._dataset, **response)

    def get(self, name: str) -> Branch:
        """Get the branch with the given name.

        Arguments:
            name: The required branch name.

        Raises:
            ResourceNotExistError: When the given name is an empty string.

        Returns:
            The :class:`.Branch` instance with the given name.

        """
        if not name:
            raise ResourceNotExistError(resource="branch", identification=name)

        response = get_branch(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            branch=name,
        )
        return Branch(self._dataset, **response)

    def list(self) -> LazyPagingList[Branch]:
        """List the information of branches.

        Returns:
            The LazyPagingList of :class:`branches<.Branch>` instances.

        """
        return LazyPagingList(self._generate, 24)

    def delete(self, name: str) -> None:
        """Delete a branch.

        Arguments:
            name: The name of the branch to be deleted.

        Raises:
            ResourceNotExistError: When the given name is an empty string.

        """
        if not name:
            raise ResourceNotExistError(resource="branch", identification=name)

        delete_branch(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            branch=name,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Operation module."""

from graviti.operation.frame import AddData, DataFrameOperation, UpdateData
from graviti.operation.sheet import CreateSheet, DeleteSheet, SheetOperation

__all__ = [
    "AddData",
    "CreateSheet",
    "DataFrameOperation",
    "DeleteSheet",
    "SheetOperation",
    "UpdateData",
]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti Series."""


from itertools import islice
from typing import Any, Iterable, List, Optional, Type, TypeVar, Union, overload

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.column.indexing import ColumnSeriesILocIndexer, ColumnSeriesLocIndexer
from graviti.dataframe.container import Container
from graviti.utility import MAX_REPR_ROWS, FileBase, PagingList

_S = TypeVar("_S", bound="Series")
_A = TypeVar("_A", bound="ArraySeries")


@pt.ContainerRegister(
    pt.binary,
    pt.boolean,
    pt.enum,
    pt.float32,
    pt.float64,
    pt.int32,
    pt.int64,
    pt.string,
)
class Series(Container):
    """One-dimensional array.

    Arguments:
        data: The data that needs to be stored in Series. Could be ndarray or Iterable.
        schema: Data type to force. If None, will be inferred from ``data``.
        name: The name to the Series.
        index: Index of the data, must have the same length as ``data``.

    Examples:
        Constructing Series from a list.

        >>> d = [1,2,3,4]
        >>> series = Series(data=d)
        >>> series
        0 1
        1 2
        2 3
        3 4

    """

    has_keys = False
    schema: pt.PortexType
    _data: PagingList
    name: Optional[str]

    def __init__(
        self,
        data: Iterable[Any],
        schema: Optional[pt.PortexType] = None,
        name: Union[str, None] = None,
    ) -> None:
        raise NotImplementedError("Not support initializing Series by __init__.")

    # @overload
    # def __getitem__(self, key: slice) -> "Series":
    #    ...

    # @overload
    # def __getitem__(self, key: int) -> Any:
    #     ...

    # @overload
    # def __getitem__(self, key: Iterable[int]) -> "Series":
    #     ...

    def __getitem__(self, key: int) -> Any:
        return self._data[key].as_py()

    @overload
    def __setitem__(self, key: slice, value: Iterable[Any]) -> None:
        ...

    @overload
    def __setitem__(self, key: int, value: Any) -> None:
        ...

    def __setitem__(
        self,
        key: Union[int, slice],
        value: Union[Iterable[Any], Any],
    ) -> None:
        pass

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        indices = list(self._get_repr_indices())
        indice_width = len(str(max(indices, default=1)))

        body = []
        body_item_width = 0
        for i in indices:
            item = self.loc[i]
            name = item._repr_folding() if hasattr(item, "_repr_folding") else str(item)
            body.append(name)
            body_item_width = max(len(name), body_item_width)

        lines = []
        for indice, value in zip(indices, body):
            lines.append(f"{indice:<{indice_width+2}}{value:<{body_item_width+2}}")
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        if self.name:
            lines.append(f"Name: {self.name}")
        return "\n".join(lines)

    def _get_repr_indices(self) -> Iterable[int]:
        return islice(range(len(self)), MAX_REPR_ROWS)

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: Iterable[int]) -> List[int]:
    #     ...

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: int) -> int:
    #     ...

    # @staticmethod
    # def _get_location_by_index(key: Union[int, Iterable[int]]) -> Union[int, List[int]]:
    #     if isinstance(key, int):
    #         return key

    #     return list(key)

    # @overload
    # def _getitem_by_location(self, key: slice) -> "Series":
    #    ...

    # @overload
    # def _getitem_by_location(self, key: int) -> Any:
    #     ...

    # @overload
    # def _getitem_by_location(self, key: Iterable[int]) -> "Series":
    #     ...

    def _getitem_by_location(self, key: int) -> Any:
        return self.__getitem__(key)

    def _extend(self, values: "Series") -> None:
        """Extend Series to itself row by row.

        Arguments:
            values: A series that needs to be extended.

        Examples:
            >>> s1 = Series.from_pyarrow(pa.array([1,2,3]))
            >>> s1
            0  1
            1  2
            2  3
            >>> s2 = Series.from_pyarrow(pa.array([4,5,6]))
            >>> s2
            0  4
            1  5
            2  6
            >>> s1.extend(s2)
            >>> s1
            0  1
            1  2
            2  3
            3  1
            4  2
            5  3

        """
        self._data.extend(values._data)  # pylint: disable=protected-access

    @property
    def iloc(self) -> ColumnSeriesILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A list or array of integers, e.g. ``[4, 3, 0]``.
        - A slice object with ints, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesILocIndexer(self)

    @property
    def loc(self) -> ColumnSeriesLocIndexer:
        """Access a group of rows and columns by indexes or a boolean array.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A list or array of indexes, e.g. ``[4, 3, 0]``.
        - A slice object with indexes, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> series = Series([1, 2, 3])
            >>> series.loc[0]
            1
            >>> df.loc[[0]]
            0    1
            dtype: int64

        """
        return ColumnSeriesLocIndexer(self)

    @classmethod
    def _from_paging(  # pylint: disable=arguments-differ
        cls: Type[_S], paging: PagingList, schema: pt.PortexType, name: Optional[str] = None
    ) -> _S:
        obj: _S = object.__new__(cls)
        obj._data = paging
        obj.schema = schema
        obj.name = name
        return obj

    @classmethod
    def _from_pyarrow(  # pylint: disable=arguments-differ
        cls: Type[_S], array: pa.Array, schema: pt.PortexType, name: Optional[str] = None
    ) -> _S:
        obj: _S = object.__new__(cls)
        obj._data = PagingList(array)
        obj.schema = schema
        obj.name = name
        return obj

    @classmethod
    def from_pyarrow(
        cls: Type[_S],
        array: pa.Array,
        schema: Optional[pt.PortexType] = None,
        name: Optional[str] = None,
    ) -> _S:
        """Instantiate a Series backed by an pyarrow array.

        Arguments:
            array: The input pyarrow array.
            schema: The schema of the series. If None, will be inferred from `array`.
            name: The name to the Series.

        Raises:
            TypeError: When the schema is mismatched with the pyarrow array type.

        Returns:
            The loaded :class:`~graviti.dataframe.column.Series` instance.

        """
        if schema is None:
            schema = pt.PortexType.from_pyarrow(array.type)
        elif not array.type.equals(schema.to_pyarrow()):
            raise TypeError("The schema is mismatched with the pyarrow array.")

        return cls._from_pyarrow(array, schema, name)

    def to_pylist(self) -> List[Any]:
        """Convert the Series to a python list.

        Returns:
            The python list representing the Series.

        """
        return self._data.to_pyarrow().to_pylist()  # type: ignore[no-any-return]

    def copy(self: _S) -> _S:
        """Get a copy of the series.

        Returns:
            A copy of the series.

        """
        obj: _S = object.__new__(self.__class__)

        obj.name = self.name
        obj.schema = self.schema.copy()
        obj._data = self._data.copy()  # pylint: disable=protected-access

        return obj


@pt.ContainerRegister(pt.array)
class ArraySeries(Series):  # pylint: disable=abstract-method
    """One-dimensional array for portex builtin type array."""

    _item_container: Type[Container]

    def __getitem__(self, key: int) -> Any:
        scalar = self._data[key]
        if not scalar.is_valid:
            return None

        return self._item_container._from_pyarrow(
            scalar.values, self.schema.to_builtin().items  # type: ignore[attr-defined]
        )

    @classmethod
    def _from_paging(  # pylint: disable=arguments-differ
        cls: Type[_A], paging: PagingList, schema: pt.PortexType, name: Optional[str] = None
    ) -> _A:
        obj = super()._from_paging(paging, schema, name)

        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[attr-defined]
        obj._item_container = builtin_schema.items.container  # pylint: disable=protected-access

        return obj

    @classmethod
    def _from_pyarrow(  # pylint: disable=arguments-differ
        cls: Type[_A], array: pa.Array, schema: pt.PortexType, name: Optional[str] = None
    ) -> _A:
        obj = super()._from_pyarrow(array, schema, name)

        builtin_schema: pt.array = schema.to_builtin()  # type: ignore[attr-defined]
        obj._item_container = builtin_schema.items.container  # pylint: disable=protected-access

        return obj

    def copy(self: _A) -> _A:
        """Get a copy of the series.

        Returns:
            A copy of the series.

        """
        obj = super().copy()
        obj._item_container = self._item_container  # pylint: disable=protected-access

        return obj


@pt.ExternalContainerRegister(
    "https://github.com/Project-OpenBytes/standard",
    "main",
    "file.Audio",
    "file.Image",
    "file.PointCloud",
    "file.PointCloudBin",
    "file.RemoteFile",
)
class FileSeries(Series):  # pylint: disable=abstract-method
    """One-dimensional array for file."""

    def __getitem__(self, key: int) -> Optional[FileBase]:
        scalar = self._data[key]
        if not scalar.is_valid:
            return None

        return FileBase.from_pyarrow(scalar)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Graviti DataFrame."""


from itertools import chain, islice, zip_longest
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

import graviti.portex as pt
from graviti.dataframe.column.series import ArraySeries, FileSeries
from graviti.dataframe.column.series import Series as ColumnSeries
from graviti.dataframe.container import Container
from graviti.dataframe.indexing import DataFrameILocIndexer, DataFrameLocIndexer
from graviti.dataframe.row.series import Series as RowSeries
from graviti.operation import AddData, DataFrameOperation
from graviti.utility import MAX_REPR_ROWS, File, PagingLists

_T = TypeVar("_T", bound="DataFrame")
_C = TypeVar("_C", bound="Container")


@pt.ContainerRegister(pt.record)
class DataFrame(Container):
    """Two-dimensional, size-mutable, potentially heterogeneous tabular data.

    Arguments:
        data: The data that needs to be stored in DataFrame.
        schema: The schema of the DataFrame. If None, will be inferred from `data`.
        columns: Column labels to use for resulting frame when data does not have them,
            defaulting to RangeIndex(0, 1, 2, ..., n). If data contains column labels,
            will perform column selection instead.

    Examples:
        Constructing DataFrame from list.

        >>> df = DataFrame(
        ...     [
        ...         {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
        ...         {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
        ...         {"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}},
        ...     ]
        ... )
        >>> df
            filename box2ds
                     x      y
        0   a.jpg    1      1
        1   b.jpg    2      2
        2   c.jpg    3      3

    """

    has_keys = True
    _columns: Dict[str, Container]
    _column_names: List[str]
    schema: pt.PortexType
    operations: List[DataFrameOperation]

    def __new__(
        cls: Type[_T],
        data: Union[List[Dict[str, Any]], _T, None] = None,
        schema: Optional[pt.PortexType] = None,
    ) -> _T:
        """Two-dimensional, size-mutable, potentially heterogeneous tabular data.

        Arguments:
            data: The data that needs to be stored in DataFrame.
            schema: The schema of the DataFrame. If None, will be inferred from `data`.

        Raises:
            ValueError: When both given schema and DataFrame data.
            ValueError: When the given data is not list.

        Returns:
            The created :class:`~graviti.dataframe.DataFrame` object.

        """
        if data is None:
            data = []

        if isinstance(data, cls):
            if schema is not None:
                raise ValueError("Only support the schema when data is not DataFrame.")

            return data.copy()

        if schema is None:
            return cls.from_pyarrow(pa.array(data))

        if isinstance(data, list):
            return cls._from_pyarrow(cls._pylist_to_pyarrow(data, schema), schema)

        raise ValueError("DataFrame only supports creating from list object now")

    @overload
    def __getitem__(self, key: str) -> Union[ColumnSeries, "DataFrame"]:  # type: ignore[misc]
        # https://github.com/python/mypy/issues/5090
        ...

    @overload
    def __getitem__(self, key: Iterable[str]) -> "DataFrame":
        ...

    def __getitem__(self, key: Union[str, Iterable[str]]) -> Container:
        if isinstance(key, str):
            return self._columns[key]

        if isinstance(key, Iterable):
            new_columns = {name: self._columns[name] for name in key}
            return self._construct(new_columns)

        raise KeyError(key)

    def __setitem__(self, key: str, value: Union[Iterable[Any], _C]) -> None:
        if isinstance(value, Container):
            schema = value.schema.copy()
            column = value.copy()
        else:
            # TODO: Need to support the case where iterable elements are subclass of Container.
            array = pa.array(value)
            schema = pt.PortexType.from_pyarrow(array.type)
            column = schema.container._from_pyarrow(array, schema)

        if len(column) != len(self):
            raise ValueError(
                f"Length of values ({len(column)}) does not match length of "
                f"self DataFrame ({len(self)})"
            )

        # TODO: Need edit the schema.
        self._columns[key] = column
        self._column_names.append(key)
        # TODO: Need add corresponding operations.

    def __repr__(self) -> str:
        flatten_header, flatten_data = self._flatten()
        header = self._get_repr_header(flatten_header)
        body = self._get_repr_body(flatten_data)
        column_widths = [len(max(column, key=len)) for column in zip_longest(*header, *body)]
        lines = [
            "".join(f"{item:<{column_widths[index]+2}}" for index, item in enumerate(line))
            for line in chain(header, body)
        ]
        if self.__len__() > MAX_REPR_ROWS:
            lines.append(f"...({self.__len__()})")
        return "\n".join(lines)

    def __len__(self) -> int:
        return self._columns[self._column_names[0]].__len__()

    @staticmethod
    def _get_repr_header(flatten_header: List[Tuple[str, ...]]) -> List[List[str]]:
        lines: List[List[str]] = []
        for names in zip_longest(*flatten_header, fillvalue=""):
            line = [""]
            pre_name = None
            upper_line = lines[-1][1:] if lines else [""]
            for name, upper_name in zip_longest(names, upper_line, fillvalue=""):
                line.append("" if name == pre_name and upper_name == "" else name)
                pre_name = name
            lines.append(line)
        return lines

    @classmethod
    def _construct(cls, columns: Dict[str, Container]) -> "DataFrame":
        obj: DataFrame = object.__new__(cls)
        obj._columns = columns
        obj._column_names = list(obj._columns.keys())
        return obj

    @classmethod
    def _from_pyarrow(cls: Type[_T], array: pa.StructArray, schema: pt.PortexType) -> _T:
        obj: _T = object.__new__(cls)
        obj.schema = schema
        obj._columns = {}
        obj._column_names = []

        # In this case schema.to_builtin always returns record.
        for key, value in schema.to_builtin().fields.items():  # type: ignore[attr-defined]
            obj._columns[key] = value.container._from_pyarrow(  # pylint: disable=protected-access
                array.field(key), schema=value
            )

            obj._column_names.append(key)

        obj.operations = [AddData(obj.copy())]
        return obj

    @classmethod
    def _from_paging(cls: Type[_T], paging: PagingLists, schema: pt.PortexType) -> _T:
        """Create DataFrame with paging lists.

        Arguments:
            paging: A dict used to initialize dataframe whose data is stored in
                lazy-loaded form.
            schema: The schema of the DataFrame.

        Returns:
            The loaded :class:`~graviti.dataframe.DataFrame` object.

        """
        obj: _T = object.__new__(cls)
        obj.schema = schema
        obj._columns = {}
        obj._column_names = []
        obj.operations = []

        # In this case schema.to_builtin always returns record.
        for key, value in schema.to_builtin().fields.items():  # type: ignore[attr-defined]
            obj._columns[key] = value.container._from_paging(  # pylint: disable=protected-access
                paging[key], schema=value
            )

            obj._column_names.append(key)

        return obj

    @classmethod
    def from_pyarrow(
        cls: Type[_T], array: pa.StructArray, schema: Optional[pt.PortexType] = None
    ) -> _T:
        """Create DataFrame with pyarrow struct array.

        Arguments:
            array: The input pyarrow struct array.
            schema: The schema of the DataFrame.

        Raises:
            TypeError: When the given schema is mismatched with the pyarrow array type.

        Returns:
            The loaded :class:`~graviti.dataframe.DataFrame` instance.

        """
        if schema is None:
            schema = pt.PortexType.from_pyarrow(array.type)
        elif not array.type.equals(schema.to_pyarrow()):
            raise TypeError("The schema is mismatched with the pyarrow array.")

        return cls._from_pyarrow(array, schema)

    def _flatten(self) -> Tuple[List[Tuple[str, ...]], List[ColumnSeries]]:
        header: List[Tuple[str, ...]] = []
        data: List[ColumnSeries] = []
        for key, value in self._columns.items():
            if isinstance(value, DataFrame):
                nested_header, nested_data = value._flatten()  # pylint: disable=protected-access
                header.extend((key, *sub_column) for sub_column in nested_header)
                data.extend(nested_data)
            elif isinstance(value, ColumnSeries):
                data.append(value)
                header.append((key,))
            else:
                raise TypeError(f"Unsupported column type: {value.__class__}")

        return header, data

    def _repr_folding(self) -> str:
        return f"{self.__class__.__name__}{self.shape}"

    def _get_repr_indices(self) -> Iterable[int]:
        return islice(range(len(self)), MAX_REPR_ROWS)

    def _get_repr_body(self, flatten_data: List[ColumnSeries]) -> List[List[str]]:
        lines = []
        for i in self._get_repr_indices():
            line: List[str] = [str(i)]
            for column in flatten_data:
                item = column[i]
                name = (
                    item._repr_folding()  # pylint: disable=protected-access
                    if isinstance(item, DataFrame)
                    else str(item)
                )
                line.append(name)
            lines.append(line)
        return lines

    def _extend(self, values: "DataFrame") -> None:
        for key, value in self._columns.items():
            value._extend(values[key])  # pylint: disable=protected-access

    @property
    def iloc(self) -> DataFrameILocIndexer:
        """Purely integer-location based indexing for selection by position.

        Allowed inputs are:

        - An integer, e.g. ``5``.
        - A list or array of integers, e.g. ``[4, 3, 0]``.
        - A slice object with ints, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the ILocIndexer.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.iloc[0]
            col1    1
            col2    3
            Name: 0, dtype: int64
            >>> df.iloc[[0]]
               col1  col2
            0     1     3

        """
        return DataFrameILocIndexer(self)

    @property
    def loc(self) -> DataFrameLocIndexer:
        """Access a group of rows and columns by indexes or a boolean array.

        Allowed inputs are:

        - A single index, e.g. ``5``.
        - A list or array of indexes, e.g. ``[4, 3, 0]``.
        - A slice object with indexes, e.g. ``1:7``.
        - A boolean array of the same length as the axis being sliced.

        Returns:
            The instance of the LocIndexer.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.loc[0]
            col1    1
            col2    3
            Name: 0, dtype: int64
            >>> df.loc[[0]]
               col1  col2
            0     1     3

        """
        return DataFrameLocIndexer(self)

    @property
    def shape(self) -> Tuple[int, int]:
        """Return a tuple representing the dimensionality of the DataFrame.

        Returns:
            Shape of the DataFrame.

        Examples:
            >>> df = DataFrame(
            ...     [
            ...         {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...         {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ...         {"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}},
            ...     ]
            ... )
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   c.jpg    3      3
            >>> df.shape
            (3, 2)

        """
        return (self.__len__(), len(self._column_names))

    @property
    def size(self) -> int:
        """Return an int representing the number of elements in this object.

        Return:
            Size of the DataFrame.

        Examples:
            >>> df = DataFrame({"col1": [1, 2], "col2": [3, 4]})
            >>> df.size
            4

        """

    # @overload
    # def _getitem_by_location(self, key: slice) -> "DataFrame":
    #    ...

    # @overload
    # def _getitem_by_location(self, key: int) -> RowSeries:
    #     ...

    # @overload
    # def _getitem_by_location(self, key: Iterable[int]) -> "DataFrame":
    #     ...

    def _getitem_by_location(self, key: int) -> RowSeries:
        indices_data = {name: self._columns[name].iloc[key] for name in self._column_names}
        return RowSeries._construct(indices_data, key)  # pylint: disable=protected-access

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: Iterable[int]) -> Iterable[int]:
    #     ...

    # @overload
    # @staticmethod
    # def _get_location_by_index(key: int) -> int:
    #     ...

    # @staticmethod
    # def _get_location_by_index(key: Union[int, Iterable[int]]) -> Union[int, Iterable[int]]:
    #     return key

    def head(self, n: int = 5) -> "DataFrame":
        """Return the first `n` rows.

        Arguments:
            n: Number of rows to select.

        Return:
            The first `n` rows.

        Examples:
            >>> df = DataFrame(
            ...     {
            ...         "animal": [
            ...             "alligator",
            ...             "bee",
            ...             "falcon",
            ...             "lion",
            ...             "monkey",
            ...             "parrot",
            ...             "shark",
            ...             "whale",
            ...             "zebra",
            ...         ]
            ...     }
            ... )
            >>> df
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot
            6      shark
            7      whale
            8      zebra

            Viewing the first `n` lines (three in this case)

            >>> df.head(3)
                  animal
            0  alligator
            1        bee
            2     falcon

            For negative values of `n`

            >>> df.head(-3)
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot

        """

    def tail(self, n: int = 5) -> "DataFrame":
        """Return the last `n` rows.

        Arguments:
            n: Number of rows to select.

        Return:
            The last `n` rows.

        Examples:
            >>> df = DataFrame(
            ...     {
            ...         "animal": [
            ...             "alligator",
            ...             "bee",
            ...             "falcon",
            ...             "lion",
            ...             "monkey",
            ...             "parrot",
            ...             "shark",
            ...             "whale",
            ...             "zebra",
            ...         ]
            ...     }
            ... )
            >>> df
                  animal
            0  alligator
            1        bee
            2     falcon
            3       lion
            4     monkey
            5     parrot
            6      shark
            7      whale
            8      zebra

            Viewing the last 5 lines

            >>> df.tail()
               animal
            4  monkey
            5  parrot
            6   shark
            7   whale
            8   zebra

            Viewing the last `n` lines (three in this case)

            >>> df.tail(3)
              animal
            6  shark
            7  whale
            8  zebra

        """

    def copy(self: _T) -> _T:
        """Get a copy of the dataframe.

        Returns:
            A copy of the dataframe.

        """
        obj: _T = object.__new__(self.__class__)

        obj.schema = self.schema.copy()

        columns = {}
        for key, value in self._columns.items():
            columns[key] = value.copy()

        # pylint: disable=protected-access
        obj._columns = columns
        obj._column_names = self._column_names.copy()
        return obj

    def sample(self, n: Optional[int] = None, axis: Optional[int] = None) -> "DataFrame":
        """Return a random sample of items from an axis of object.

        Arguments:
            n: Number of items from axis to return.
            axis: {0 or `index`, 1 or `columns`, None}
                Axis to sample. Accepts axis number or name. Default is stat axis
                for given data type (0 for Series and DataFrames).

        Return:
            A new object of same type as caller containing `n` items randomly
            sampled from the caller object.

        """

    def info(self) -> None:
        """Print a concise summary of a DataFrame."""

    def extend(self, values: Union[Iterable[Dict[str, Any]], "DataFrame"]) -> None:
        """Extend Sequence object or DataFrame to itself row by row.

        Arguments:
            values: A sequence object or DataFrame.

        Raises:
            TypeError: When the given Dataframe mismatched with the self schema.

        Examples:
            >>> df = DataFrame([
            ...     {"filename": "a.jpg", "box2ds": {"x": 1, "y": 1}},
            ...     {"filename": "b.jpg", "box2ds": {"x": 2, "y": 2}},
            ... ])

            Extended by another list.

            >>> df.extend([{"filename": "c.jpg", "box2ds": {"x": 3, "y": 3}}])
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   c.jpg    3      3

            Extended by another DataFrame.

            >>> df2 = DataFrame([{"filename": "d.jpg", "box2ds": {"x": 4 "y": 4}}])
            >>> df.extend(df2)
            >>> df
                filename box2ds
                         x      y
            0   a.jpg    1      1
            1   b.jpg    2      2
            2   d.jpg    4      4

        """
        if not isinstance(values, DataFrame):
            values = DataFrame._from_pyarrow(  # pylint: disable=protected-access
                self._pylist_to_pyarrow(values, self.schema), self.schema
            )
        elif not self.schema.to_pyarrow().equals(values.schema.to_pyarrow()):
            raise TypeError("The schema of the given DataFrame is mismatched.")
        else:
            values = values.copy()

        self._extend(values)
        self.operations.append(AddData(values))

    @staticmethod
    def _get_process(value: Any, schema: pt.PortexType) -> Tuple[Callable[[Any], Any], bool]:
        container = schema.container
        if container == DataFrame:
            return DataFrame._process_record(value, schema)  # type: ignore[arg-type]
        if container == ArraySeries:
            return DataFrame._process_array(value, schema.items)  # type: ignore[attr-defined]
        if container == FileSeries:
            return DataFrame._process_file(value)

        return lambda x: x, False

    @staticmethod
    def _process_array(
        values: Iterable[Any], schema: pt.PortexType
    ) -> Tuple[Callable[[Any], Any], bool]:
        if isinstance(values, DataFrame):
            return lambda x: x.to_pylist(), True

        for value in values:
            if value is None:
                continue

            processor, need_process = DataFrame._get_process(value, schema)
            return lambda x: list(map(processor, x)), need_process
        return lambda x: x, False

    @staticmethod
    def _process_file(values: Iterable[Any]) -> Tuple[Callable[[Any], Any], bool]:
        if isinstance(values, File):
            return lambda x: x.to_pyobj(), True

        return lambda x: x, False

    @staticmethod
    def _process_record(
        values: Dict[str, Any], schema: pt.record
    ) -> Tuple[Callable[[Any], Any], bool]:
        processors: Dict[str, Callable[[Any], Any]] = {}
        need_process, sub_need_process = False, False
        fields = schema.to_builtin().fields
        for name, field in fields.items():
            item = values[name]
            if item is None:
                processors[name] = lambda x: x
            else:
                processors[name], sub_need_process = DataFrame._get_process(item, field)
                need_process = need_process or sub_need_process

        return lambda x: {k: processors[k](v) for k, v in x.items()}, need_process

    @staticmethod
    def _pylist_to_pyarrow(values: Iterable[Any], schema: pt.PortexType) -> pa.StructArray:

        # In this case schema.to_builtin always returns record.
        processor, need_process = DataFrame._process_array(values, schema)

        if not need_process:
            return pa.array(values, schema.to_pyarrow())

        return pa.array(processor(values), schema.to_pyarrow())

    def to_pylist(self) -> List[Dict[str, Any]]:
        """Convert the DataFrame to a python list.

        Returns:
            The python list representing the DataFrame.

        """
        return [
            dict(zip(self._column_names, values))
            for values in zip(*(column.to_pylist() for column in self._columns.values()))
        ]



#!/usr/bin/env python3

# pylint: disable-all
# type: ignore
# flake8: noqa

"""Code converting PyArrow schema to Avro Schema."""

from graviti.portex.base import PortexType
from graviti.portex.builtin import (
    array,
    binary,
    boolean,
    enum,
    float32,
    float64,
    int32,
    int64,
    record,
    string,
    tensor,
)

_REMOTE_FILE_FIELD_NAMES = {"checksum", "url"}

_PRIMITIVE_TYPES = {
    int32: "int",
    int64: "int",
    float32: "float",
    float64: "double",
    boolean: "boolean",
    string: "string",
    binary: "bytes",
}


class AvroSchema:
    def __init__(self):
        pass

    def to_json(self):
        return {}


class AvroField:
    def __init__(
        self,
        name_registry,
        typ: AvroSchema,
        name,
        index,
        optional=False,
        has_default=False,
        default=None,
    ):
        self._type = typ
        self._name = name
        self._index = index
        self._optional = optional
        self._has_default = has_default
        self._default = default

    def to_json(self):
        if self._optional:
            result = {
                "name": self._name,
                "type": [
                    "null",
                    self._type.to_json(),
                ],
            }
        else:
            result = {
                "name": self._name,
                "type": self._type.to_json(),
            }
        if self._has_default:
            result["default"] = self._default

        return result


class AvroPrimitiveSchema(AvroSchema):
    def __init__(self, typ, has_default=False, default=None):
        super().__init__()
        self._default = default
        self._has_default = has_default
        self._type = typ

    def to_json(self):
        if self._has_default:
            return {"type": self._type, "default": self._default}
        else:
            return self._type


class AvroRecordSchema(AvroSchema):
    def __init__(self, name_registry, name, namespace, fields: [], aliases=None):
        super().__init__()
        self._name = name
        self._namespace = namespace
        self._fields = fields

        self._aliases = aliases
        if self._aliases is None:
            self._aliases = []

    def to_json(self):
        return {
            "type": "record",
            "name": self._name,
            "namespace": self._namespace,
            "aliases": self._aliases,
            "fields": [field.to_json() for field in self._fields],
        }


class AvroArraySchema(AvroSchema):
    def __init__(self, items: AvroSchema):
        super().__init__()
        self._items = items

    def to_json(self):
        return {
            "type": "array",
            "items": self._items.to_json(),
            "default": [],
        }


class AvroFixedArraySchema(AvroSchema):
    def __init__(self, items: AvroSchema, shape):
        super().__init__()
        self._items = items
        self._shape = shape

    def to_json(self):
        return {
            "type": "array",
            "items": self._items.to_json(),
            "shape": self._shape,
            "default": [],
        }


class AvroEnumSchema(AvroSchema):
    def __init__(self, namespace, name, symbols):
        super().__init__()
        self._namespace = namespace
        self._name = name
        self._symbols = symbols

    def to_json(self):
        return {
            "type": "enum",
            "namespace": self._namespace,
            "name": self._name,
            "symbols": self._symbols,
        }


def _on_list(names, namespace, name, _pa_ist: array) -> AvroArraySchema:
    sub_namespace = f"{namespace}.{name}.items"
    sub_name = "items"
    items = _on_type(names, sub_namespace, sub_name, 0, _pa_ist.items.to_builtin())
    return AvroArraySchema(items=items)


def _on_primitive(portex_type: PortexType) -> AvroSchema:
    try:
        return AvroPrimitiveSchema(typ=_PRIMITIVE_TYPES[type(portex_type)])
    except KeyError:
        raise Exception(f"unsupported type {portex_type}") from None


def _on_struct(names, namespace, name, _struct: record) -> AvroRecordSchema:
    avro_record_fields = list()

    fields = _struct.fields
    num_fields = len(fields)
    index_range = range(num_fields)

    # remove "url" field in avro schema
    if num_fields == 2 and set(fields) == _REMOTE_FILE_FIELD_NAMES:
        index_range = [list(fields).index("checksum")]

    for i in index_range:
        sub_name = list(fields.keys())[i]
        sub_type = fields[i].to_builtin()
        sub_namespace = f"{namespace}.{sub_name}"
        sub_schema = _on_type(names, sub_namespace, sub_name, i, sub_type)
        avro_record_field = AvroField(
            typ=sub_schema,
            name=sub_name,
            index=i,
            has_default=False,
            name_registry=names,
            optional=sub_type.nullable,
        )
        avro_record_fields.append(avro_record_field)

    return AvroRecordSchema(
        name=name, namespace=namespace, fields=avro_record_fields, name_registry=names
    )


def _on_fixed_shape_list(names, namespace, name, _pa_ist: tensor) -> AvroFixedArraySchema:
    sub_namespace = f"{namespace}.{name}.items"
    sub_name = "items"
    items = _on_type(names, sub_namespace, sub_name, 0, _pa_ist.items.to_builtin())
    return AvroFixedArraySchema(items=items, shape=_pa_ist.shape)


def _on_enum(name_registry, namespace, name, _filed: enum) -> AvroSchema:
    return AvroEnumSchema(namespace, name, _filed.values)


def _on_type(names, namespace, name, _type_index, _portex_type):
    typ = type(_portex_type)
    if typ in _COMPLEX_TYPES_PROCESSERS:
        return _COMPLEX_TYPES_PROCESSERS[typ](names, namespace, name, _portex_type)

    return _on_primitive(_portex_type)


def convert_portex_schema_to_avro(_schema: record):
    avro_record_fields = list()
    namespace = "cn.graviti.portex"
    name = "root"
    names = []

    for i, (sub_name, typ) in enumerate(_schema.to_builtin().fields.items()):
        sub_namespace = f"{namespace}.{name}"
        sub_schema = _on_type(names, sub_namespace, sub_name, i, typ.to_builtin())
        avro_record_field = AvroField(
            typ=sub_schema,
            name=sub_name,
            index=i,
            has_default=False,
            name_registry=names,
            optional=typ.nullable,
        )

        avro_record_fields.append(avro_record_field)

    avro_schema = AvroRecordSchema(
        name=name, namespace=namespace, fields=avro_record_fields, name_registry=names
    )
    return avro_schema.to_json()


_COMPLEX_TYPES_PROCESSERS = {
    record: _on_struct,
    array: _on_list,
    enum: _on_enum,
    tensor: _on_fixed_shape_list,
}



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Lazy list related class."""

from bisect import bisect_right
from itertools import chain
from math import ceil
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

from graviti.utility.pyarrow import GravitiExtension
from graviti.utility.repr import ReprMixin, ReprType

_T = TypeVar("_T")


class LazyList(Sequence[_T], ReprMixin):
    """LazyList is a lazy load list which follows Sequence protocol.

    Arguments:
        total_count: The total count of the elements in the lazy list.
        limit: The size of each lazy load page.
        fetcher: A callable object to fetch the data and load it to the lazy list.
        extractor: A callable object to make the source data to an iterable object.
        dtype: The pyarrow data type of the elements in the lazy list.

    Attributes:
        pages: A list of pyarrow arrays that contains the data in the lazy list.

    """

    _repr_type = ReprType.SEQUENCE

    def __init__(
        self,
        total_count: int,
        limit: int,
        fetcher: Callable[[int], None],
        extractor: Callable[[Any], Iterable[Any]],
        *,
        dtype: Optional[pa.DataType] = None,
    ) -> None:
        self.pages: List[Union[LazyPage[_T], pa.Array]] = [
            LazyPage(i, fetcher, self) for i in range(ceil(total_count / limit))
        ]
        self._limit = limit
        self._total_count = total_count
        self._extractor = extractor
        self._dtype = dtype

    def __len__(self) -> int:
        return self._total_count

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> pa.Array:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[_T, pa.Array]:
        if isinstance(index, slice):
            return pa.array([self._getitem(i) for i in range(*index.indices(self._total_count))])

        index = index if index >= 0 else self._total_count + index
        return self._getitem(index)

    def __iter__(self) -> Iterator[_T]:
        return chain(*self.pages)

    def _getitem(self, index: int) -> _T:
        """Get item from the lazy list.

        Arguments:
            index: The index of the item which needs to be non-negative.

        Returns:
            The item at the index position.

        """
        div, mod = divmod(index, self._limit)
        return self.pages[div][mod]

    def _create_array(self, values: Iterable[_T]) -> pa.Array:
        if isinstance(self._dtype, GravitiExtension):
            storage_array = pa.array(values, type=self._dtype.storage_type)
            return pa.ExtensionArray.from_storage(self._dtype, storage_array)

        return pa.array(values, type=self._dtype)

    def update(self, pos: int, data: Any) -> None:
        """Update one page by the given data.

        Arguments:
            pos: The page number.
            data: The source data which needs to be input to the extractor.

        """
        self.pages[pos] = self._create_array(self._extractor(data))


class MutableLazyList(LazyList[_T]):
    """MutableLazyList is a lazy load list which follows Sequence protocol.

    It supports extend method to add items into the list.

    Arguments:
        total_count: The total count of the elements in the lazy list.
        limit: The size of each lazy load page.
        fetcher: A callable object to fetch the data and load it to the lazy list.
        extractor: A callable object to make the source data to an iterable object.
        dtype: The pyarrow data type of the elements in the lazy list.

    Attributes:
        pages: A list of pyarrow arrays that contains the data in the lazy list.

    """

    _repr_type = ReprType.SEQUENCE

    def __init__(
        self,
        total_count: int,
        limit: int,
        fetcher: Callable[[int], None],
        extractor: Callable[[Any], Iterable[Any]],
        *,
        dtype: Optional[pa.DataType] = None,
    ) -> None:
        super().__init__(total_count, limit, fetcher, extractor, dtype=dtype)
        self._offsets = list(range(0, total_count, limit))

    def _getitem(self, index: int) -> _T:
        """Get item from the lazy list.

        Arguments:
            index: The index of the item which needs to be non-negative.

        Returns:
            The item at the index position.

        """
        page_number = bisect_right(self._offsets, index) - 1
        offset = self._offsets[page_number]
        return self.pages[page_number][index - offset]

    def extend(self, values: Iterable[_T]) -> None:
        """Extend mutable sequence by appending elements from the iterable.

        Arguments:
            values: Elements to be Extended into the mutable sequence.

        """
        page = self._create_array(values)
        self._offsets.append(self._total_count)
        self._total_count += len(page)
        self.pages.append(page)


class LazyPage(Generic[_T]):
    """LazyPage is a placeholder of the pages in the lazy list when the page is not loaded yet.

    Arguments:
        pos: The page number.
        fetcher: A callable object to fetch the data and load it to the lazy list.
        parent: The parent lazy list.

    """

    def __init__(self, pos: int, fetcher: Callable[[int], None], parent: LazyList[_T]) -> None:
        self._pos = pos
        self._fetcher = fetcher
        self._parent = parent

    def __getitem__(self, index: int) -> _T:
        self._fetcher(self._pos)
        return self._parent.pages[self._pos].__getitem__(index)

    def __iter__(self) -> Iterator[_T]:
        self._fetcher(self._pos)
        return self._parent.pages[self._pos].__iter__()


class LazyFactory:
    """LazyFactory is a factory for creating lazy lists.

    Arguments:
        total_count: The total count of the elements in the lazy lists.
        limit: The size of each lazy load page.
        getter: A callable object to get the source data.

    Examples:
        >>> TOTAL_COUNT = 1000
        >>> def getter(offset: int, limit: int) -> Dict[str, Any]:
        ...     stop = min(offset + limit, TOTAL_COUNT)
        ...     data = [
        ...         {
        ...             "remotePath": f"{i:06}.jpg",
        ...             "label": {"CLASSIFICATION": {"category": "cat" if i % 2 else "dog"}},
        ...         }
        ...         for i in range(offset, stop)
        ...     ]
        ...
        ...     return {
        ...         "data": data,
        ...         "offset": offset,
        ...         "recordSize": len(data),
        ...         "totalCount": TOTAL_COUNT,
        ...     }
        >>> factory = LazyFactory(TOTAL_COUNT, 128, getter)
        >>> paths = factory.create_list(
        ...     lambda data: (item["remotePath"] for item in data["data"]), dtype="<U10"
        ... )
        >>> categories = factory.create_list(
        ...     lambda data: (item["label"]["CLASSIFICATION"]["category"] for item in data["data"]),
        ...     dtype="<U3",
        ... )
        >>> paths
        LazyList [
          '000000.jpg',
          '000001.jpg',
          '000002.jpg',
          '000003.jpg',
          '000004.jpg',
          '000005.jpg',
          '000006.jpg',
          '000007.jpg',
          '000008.jpg',
          '000009.jpg',
          '000010.jpg',
          '000011.jpg',
          '000012.jpg',
          '000013.jpg',
          ... (985 items are folded),
          '000999.jpg'
        ]
        >>> categories
        LazyList [
          'dog',
          'cat',
          'dog',
          'cat',
          'dog',
          'cat',
          'dog',
          'cat',
          'dog',
          'cat',
          'dog',
          'cat',
          'dog',
          'cat',
          ... (985 items are folded),
          'cat'
        ]

    """

    def __init__(self, total_count: int, limit: int, getter: Callable[[int, int], Any]) -> None:
        self._getter = getter
        self._total_count = total_count
        self._limit = limit
        self._lists: List[LazyList[Any]] = []

    def create_list(
        self, extractor: Callable[[Any], Iterable[Any]], dtype: Optional[pa.DataType] = None
    ) -> LazyList[Any]:
        """Create a lazy list from the factory.

        Arguments:
            extractor: A callable object to make the source data to an iterable object.
            dtype: The pyarrow data type of the elements in the lazy list.

        Returns:
            A lazy list created by the given extractor and dtype.

        """
        lazy_list: LazyList[Any] = LazyList(
            self._total_count, self._limit, self.fetch, extractor, dtype=dtype
        )
        self._lists.append(lazy_list)
        return lazy_list

    def create_mutable_list(
        self, extractor: Callable[[Any], Iterable[Any]], dtype: Optional[pa.DataType] = None
    ) -> LazyList[Any]:
        """Create a mutable lazy list from the factory.

        Arguments:
            extractor: A callable object to make the source data to an iterable object.
            dtype: The pyarrow data type of the elements in the lazy list.

        Returns:
            A lazy list created by the given extractor and dtype.

        """
        lazy_list: LazyList[Any] = MutableLazyList(
            self._total_count, self._limit, self.fetch, extractor, dtype=dtype
        )
        self._lists.append(lazy_list)
        return lazy_list

    def fetch(self, pos: int) -> None:
        """Fetch the source data and load the data to all lazy lists.

        Arguments:
            pos: The page number.

        """
        data = self._getter(self._limit * pos, self._limit)
        for lazy_list in self._lists:
            lazy_list.update(pos, data)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Dataset and DatasetManager."""

from typing import Any, Dict, Generator, Optional, Tuple, Type, TypeVar

from tensorbay.utility import AttrsMixin, attr, common_loads

from graviti.dataframe import DataFrame
from graviti.exception import ResourceNotExistError
from graviti.manager.branch import BranchManager
from graviti.manager.commit import Commit, CommitManager
from graviti.manager.draft import DraftManager
from graviti.manager.lazy import LazyPagingList
from graviti.manager.tag import TagManager
from graviti.openapi import (
    create_dataset,
    delete_dataset,
    get_dataset,
    list_datasets,
    update_dataset,
)
from graviti.utility import ReprMixin, ReprType, UserMutableMapping


class Dataset(  # pylint: disable=too-many-instance-attributes
    UserMutableMapping[str, DataFrame], AttrsMixin, ReprMixin
):
    """This class defines the basic concept of the dataset on Graviti.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        dataset_id: Dataset ID.
        name: The name of the dataset, unique for a user.
        alias: Dataset alias.
        default_branch: The default branch of dataset.
        commit_id: The commit ID.
        created_at: The time when the dataset was created.
        updated_at: The time when the dataset was last modified.
        owner: The owner of the dataset.
        is_public: Whether the dataset is public.
        config: The config of dataset.

    """

    _T = TypeVar("_T", bound="Dataset")

    _repr_type = ReprType.INSTANCE

    _repr_attrs: Tuple[str, ...] = (
        "alias",
        "default_branch",
        "created_at",
        "updated_at",
        "is_public",
        "config",
        "branch",
    )

    _dataset_id: str = attr(key="id")

    access_key: str = attr()
    url: str = attr()
    name: str = attr()
    alias: str = attr()
    default_branch: str = attr()
    created_at: str = attr()
    updated_at: str = attr()
    owner: str = attr()
    is_public: bool = attr()
    config: str = attr()

    def __init__(
        self,
        access_key: str,
        url: str,
        dataset_id: str,
        name: str,
        *,
        alias: str,
        default_branch: str,
        commit_id: str,
        created_at: str,
        updated_at: str,
        owner: str,
        is_public: bool,
        config: str,
    ) -> None:
        self.access_key = access_key
        self.url = url
        self._dataset_id = dataset_id
        self.name = name
        self.alias = alias
        self.default_branch = default_branch
        self.created_at = created_at
        self.updated_at = updated_at
        self.owner = owner
        self.is_public = is_public
        self.config = config
        self._data: Commit = self.branches.get(default_branch)
        if commit_id != self._data.commit_id:
            self._data = self.commits.get(commit_id)

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.owner}/{self.name}")'

    @classmethod
    def from_pyobj(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Create a :class:`Dataset` instance from python dict.

        Arguments:
            contents: A python dict containing all the information of the dataset::

                    {
                        "access_key": <str>
                        "url": <str>
                        "id": <str>
                        "name": <str>
                        "alias": <str>
                        "default_branch": <str>
                        "commit_id": <str>
                        "created_at": <str>
                        "updated_at": <str>
                        "owner": <str>
                        "is_public": <bool>
                        "config": <str>
                    }

        Returns:
            A :class:`Dataset` instance created from the input python dict.

        """
        dataset = common_loads(cls, contents)
        default_branch = dataset.default_branch
        dataset._data = dataset.branches.get(default_branch)
        return dataset

    @property
    def HEAD(self) -> Commit:  # pylint: disable=invalid-name
        """Return the current branch or commit.

        Returns:
            The current branch or commit.

        """
        return self._data

    @property
    def branches(self) -> BranchManager:
        """Get class :class:`~graviti.manager.branch.BranchManager` instance.

        Returns:
            Required :class:`~graviti.manager.branch.BranchManager` instance.

        """
        return BranchManager(self)

    @property
    def drafts(self) -> DraftManager:
        """Get class :class:`~graviti.manager.draft.DraftManager` instance.

        Returns:
            Required :class:`~graviti.manager.draft.DraftManager` instance.

        """
        return DraftManager(self)

    @property
    def commits(self) -> CommitManager:
        """Get class :class:`~graviti.manager.commit.CommitManager` instance.

        Returns:
            Required :class:`~graviti.manager.commit.CommitManager` instance.

        """
        return CommitManager(self)

    @property
    def tags(self) -> TagManager:
        """Get class :class:`~graviti.manager.tag.TagManager` instance.

        Returns:
            Required :class:`~graviti.manager.tag.TagManager` instance.

        """
        return TagManager(self)

    def checkout(self, revision: str) -> None:
        """Checkout to a commit.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch, or the tag.

        """
        try:
            self._data = self.branches.get(revision)
        except ResourceNotExistError:
            self._data = self.commits.get(revision)

    def edit(
        self,
        *,
        name: Optional[str] = None,
        alias: Optional[str] = None,
        default_branch: Optional[str] = None,
    ) -> None:
        """Update the meta data of the dataset.

        Arguments:
            name: The new name of the dataset.
            alias: The new alias of the dataset.
            default_branch: The new default branch of the dataset.

        """
        response = update_dataset(
            self.access_key,
            self.url,
            self.owner,
            self.name,
            name=name,
            alias=alias,
            default_branch=default_branch,
        )
        self.name = response["name"]
        self.alias = response["alias"]
        self.default_branch = response["default_branch"]
        self.updated_at = response["updated_at"]

    def search(self, sheet: str, criteria: Dict[str, Any]) -> DataFrame:
        """Create a search.

        Arguments:
            sheet: The sheet name.
            criteria: The criteria of search.

        Returns:
            The created :class:`~graviti.dataframe.DataFrame` instance.

        """
        return self.HEAD.search(sheet, criteria)


class DatasetManager:
    """This class defines the operations on the dataset on Graviti.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.

    """

    def __init__(self, access_key: str, url: str, owner: str) -> None:
        self.access_key = access_key
        self.url = url
        self.owner = owner

    def _generate(self, offset: int, limit: int) -> Generator[Dataset, None, int]:
        arguments = {"access_key": self.access_key, "url": self.url}
        response = list_datasets(**arguments, limit=limit, offset=offset)

        for item in response["datasets"]:
            item.update(arguments)
            yield Dataset.from_pyobj(item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(
        self,
        name: str,
        alias: str = "",
        config: Optional[str] = None,
    ) -> Dataset:
        """Create a Graviti dataset with given name.

        Arguments:
            name: The name of the dataset, unique for a user.
            alias: Alias of the dataset, default is "".
            config: The auth storage config name.

        Returns:
            The created :class:`~graviti.manager.dataset.Dataset` instance.

        """
        arguments = {"access_key": self.access_key, "url": self.url}
        response = create_dataset(
            **arguments,
            name=name,
            alias=alias,
            config=config,
            with_draft=False,
        )
        response.update(arguments)

        return Dataset.from_pyobj(response)

    def get(self, dataset: str) -> Dataset:
        """Get a Graviti dataset with given name.

        Arguments:
            dataset: The name of the dataset, unique for a user.

        Returns:
            The requested :class:`~graviti.manager.dataset.Dataset` instance.

        Raises:
            ResourceNotExistError: When the required dataset does not exist.

        """
        if not dataset:
            raise ResourceNotExistError(resource="dataset", identification=dataset)

        arguments = {"access_key": self.access_key, "url": self.url}
        response = get_dataset(**arguments, owner=self.owner, dataset=dataset)
        response.update(arguments)

        return Dataset.from_pyobj(response)

    def list(self) -> LazyPagingList[Dataset]:
        """List Graviti datasets.

        Returns:
            The LazyPagingList of :class:`~graviti.manager.dataset.Dataset` instances.

        """
        return LazyPagingList(self._generate, 24)

    def delete(self, name: str) -> None:
        """Delete a Graviti dataset with given name.

        Arguments:
            name: The name of the dataset, unique for a user.

        """
        delete_dataset(self.access_key, self.url, self.owner, name)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Dataframe module."""

from graviti.dataframe.container import Container
from graviti.dataframe.frame import DataFrame

__all__ = ["DataFrame", "Container"]



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Template base class."""


from inspect import Signature
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Type

import pyarrow as pa

import graviti.portex.ptype as PTYPE
from graviti.portex.base import PortexType
from graviti.portex.builtin import PortexBuiltinType
from graviti.portex.factory import Factory, type_factory_creator
from graviti.portex.package import ExternalPackage, Imports, Package, packages
from graviti.portex.param import Param, Params
from graviti.portex.register import ExternalContainerRegister

EXTERNAL_TYPE_TO_CONTAINER = ExternalContainerRegister.EXTERNAL_TYPE_TO_CONTAINER


class PortexExternalType(PortexType):  # pylint: disable=abstract-method
    """The base class of Portex external type."""

    _signature: ClassVar[Signature]

    package: ClassVar[ExternalPackage]
    factory: ClassVar[Factory]

    def __init_subclass__(cls) -> None:
        cls._signature = cls.params.get_signature()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        bound_arguments = self._signature.bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        arguments = bound_arguments.arguments

        for key, value in arguments.items():
            arguments[key] = self.params[key].check(value)

        self.__dict__.update(arguments)

        if not hasattr(self.__class__, "container"):
            self.container = self.internal_type.container

    @property
    def internal_type(self) -> PortexType:
        """Get the internal type of the PortexExternalType.

        Returns:
            The internal type of the PortexExternalType.

        """
        arguments = {name: getattr(self, name) for name in self.params}
        return self.factory(**arguments)  # type: ignore[no-any-return]

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return self.internal_type.to_pyarrow()

    def to_builtin(self) -> PortexBuiltinType:
        """Expand the top level of the Portex external type to Portex builtin type.

        Returns:
            The expanded Portex builtin type.

        """
        return self.internal_type.to_builtin()  # type: ignore[attr-defined, no-any-return]

    def get_keys(self, type_name: Optional[str] = None) -> List[Tuple[str, ...]]:
        """Get the keys to locate all data, or only get keys of one type if type_name is given.

        Arguments:
            type_name: The name of the target PortexType.

        Returns:
            A list of keys to locate the data.

        """
        if not self.container.has_keys:
            return []

        return self.to_builtin().get_keys(type_name)


def template(
    name: str, content: Dict[str, Any], package: Package[Any] = packages.locals
) -> Type[PortexExternalType]:
    """Generate a Portex external type according to the input template.

    Arguments:
        name: The class name of the Portex external type.
        content: A dict indicates a Portex template.
        package: The package this template belongs to.

    Returns:
        A subclass of ``PortexExternalType``.

    Examples:
        >>> import graviti.portex as pt
        >>> from graviti.portex.template import template
        >>>
        >>> vector_template = {
        ...     "type": "template",
        ...     "parameters": [
        ...         {
        ...             "name": "coords",
        ...             "default": {"type": "int32"},
        ...         },
        ...         {
        ...             "name": "labels",
        ...             "default": None,
        ...         },
        ...     ],
        ...     "declaration": {
        ...         "type": "record",
        ...         "fields": [
        ...             {
        ...                 "name": "x",
        ...                 "+": "$coords",
        ...             },
        ...             {
        ...                 "name": "y",
        ...                 "+": "$coords",
        ...             },
        ...             {
        ...                 "name": "label",
        ...                 "exist_if": "$labels",
        ...                 "type": "enum",
        ...                 "values": "$labels",
        ...             },
        ...         ],
        ...     },
        ... }
        >>> Vector = template("Vector", vector_template)
        >>> vector2d_int = Vector()
        >>> vector2d_int
        Vector()
        >>> vector2d_int.internal_type
        record(
          fields={
            'x': int32(),
            'y': int32(),
          },
        )
        >>>
        >>> vector2d_float = Vector(pt.float32())
        >>> vector2d_float
        Vector(
          coords=float32(),
        )
        >>> vector2d_float.internal_type
        record(
          fields={
            'x': float32(),
            'y': float32(),
          },
        )
        >>>
        >>> labeled_vector = Vector(labels=["visble", "occluded"])
        >>> labeled_vector
        Vector(
          labels=['visble', 'occluded'],
        )
        >>> labeled_vector.internal_type
        record(
          fields={
            'x': int32(),
            'y': int32(),
            'label': enum(
              values=['visble', 'occluded'],
            ),
          },
        )

    """
    params_pyobj = content.get("parameters", [])
    decl = content["declaration"]

    imports = Imports.from_pyobj(content.get("imports", []))
    imports.update_base_package(package)

    factory = type_factory_creator(decl, imports)

    keys = factory.keys
    params = Params.from_pyobj(params_pyobj)

    for key, value in params.items():
        value.ptype = keys.get(key, PTYPE.Any)

    params.add(Param("nullable", False, ptype=PTYPE.Boolean))

    type_: Type[PortexExternalType] = type(
        name,
        (PortexExternalType,),
        {"params": params, "factory": factory},
    )

    if isinstance(package, ExternalPackage):
        container = EXTERNAL_TYPE_TO_CONTAINER.get((package.url, package.revision, name))
        if container:
            type_.container = container

    package[name] = type_

    return type_



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The table-structured data container related classes."""


from typing import Any, ClassVar, List, Type, TypeVar

import pyarrow as pa

from graviti.portex import PortexType

_T = TypeVar("_T", bound="Container")


class Container:
    """The base class for the table-structured data container."""

    has_keys: ClassVar[bool]
    schema: PortexType

    def __len__(self) -> int:
        raise NotImplementedError

    @classmethod
    def _from_paging(cls: Type[_T], paging: pa.Array, schema: PortexType) -> _T:
        raise NotImplementedError

    @classmethod
    def _from_pyarrow(cls: Type[_T], array: pa.Array, schema: PortexType) -> _T:
        raise NotImplementedError

    def _extend(self: _T, values: _T) -> None:
        raise NotImplementedError

    # TODO: Defines a base indexer for the iloc return type.
    @property
    def iloc(self) -> Any:
        """Purely integer-location based indexing for selection by position.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def to_pylist(self) -> List[Any]:
        """Convert the container to a python list.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    def copy(self: _T) -> _T:
        """Get a copy of the container.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Commit and CommitManager."""

from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Tuple

from tensorbay.utility import AttrsMixin, attr

from graviti.dataframe import DataFrame
from graviti.manager.lazy import LazyPagingList
from graviti.manager.sheets import Sheets
from graviti.openapi import (
    create_search,
    get_commit_sheet,
    get_revision,
    list_commit_data,
    list_commit_sheets,
    list_commits,
)
from graviti.portex import PortexType
from graviti.utility import LazyFactory

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset

ROOT_COMMIT_ID = "00000000000000000000000000000000"


class Commit(Sheets, AttrsMixin):
    """This class defines the structure of a commit.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.
        committed_at: The time when the draft is committed.

    """

    _repr_attrs: Tuple[str, ...] = ("parent_commit_id", "title", "committer", "committed_at")

    commit_id: str = attr()
    parent_commit_id: str = attr(default="")
    title: str = attr()
    description: str = attr(default="")
    committer: str = attr()
    committed_at: str = attr()

    def __init__(  # pylint: disable=too-many-arguments
        self,
        dataset: "Dataset",
        commit_id: str,
        parent_commit_id: str,
        title: str,
        description: str,
        committer: str,
        committed_at: str,
    ) -> None:
        self._dataset = dataset
        self.commit_id = commit_id
        self.parent_commit_id = parent_commit_id
        self.title = title
        self.description = description
        self.committer = committer
        self.committed_at = committed_at

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.commit_id}")'

    def _list_data(self, offset: int, limit: int, sheet_name: str) -> Dict[str, Any]:
        return list_commit_data(  # type: ignore[no-any-return]
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            commit_id=self.commit_id,
            sheet=sheet_name,
            offset=offset,
            limit=limit,
        )["data"]

    def _list_sheets(self) -> Dict[str, Any]:
        return list_commit_sheets(
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            commit_id=self.commit_id,
        )

    def _get_sheet(self, sheet_name: str) -> Dict[str, Any]:
        return get_commit_sheet(
            access_key=self._dataset.access_key,
            url=self._dataset.url,
            owner=self._dataset.owner,
            dataset=self._dataset.name,
            commit_id=self.commit_id,
            sheet=sheet_name,
            with_record_count=True,
        )

    def to_pyobj(self) -> Dict[str, str]:
        """Dump the instance to a python dict.

        Returns:
            A python dict containing all the information of the commit::

                {
                    "commit_id": <str>
                    "parent_commit_id": <str>
                    "title": <str>
                    "description": <str>
                    "committer":  <str>
                    "committed_at": <str>
                }

        """
        return self._dumps()

    def search(self, sheet: str, criteria: Dict[str, Any]) -> DataFrame:
        """Create a search.

        Arguments:
            sheet: The sheet name.
            criteria: The criteria of search.

        Returns:
            The created :class:`~graviti.dataframe.DataFrame` instance.

        """

        def _getter(offset: int, limit: int) -> Dict[str, Any]:
            return create_search(
                self._dataset.access_key,
                self._dataset.url,
                self._dataset.owner,
                self._dataset.name,
                commit_id=self.commit_id,
                sheet=sheet,
                criteria=criteria,
                offset=offset,
                limit=limit,
            )

        total_count = _getter(0, 1)["total_count"]
        schema = PortexType.from_yaml(self._get_sheet(sheet)["schema"])

        factory = LazyFactory(
            total_count,
            128,
            lambda offset, limit: _getter(offset, limit)["data"],
            schema.to_pyarrow(),
        )
        paging_lists = factory.create_lists(schema.get_keys())
        return DataFrame._from_paging(paging_lists, schema)  # pylint: disable=protected-access


class NamedCommit(Commit):  # pylint: disable=too-many-instance-attributes
    """This class defines the structure of a named commit.

    :class:`NamedCommit` is the base class of :class:`~graviti.manager.branch.Branch`
    and :class:`~graviti.manager.tag.Tag`.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the named commit.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.
        committed_at: The time when the draft is committed.

    """

    _repr_attrs = ("commit_id",) + Commit._repr_attrs

    name: str = attr()

    def __init__(  # pylint: disable=too-many-arguments, super-init-not-called
        self,
        dataset: "Dataset",
        name: str,
        commit_id: str,
        parent_commit_id: str,
        title: str,
        description: str,
        committer: str,
        committed_at: str,
    ) -> None:
        self._dataset = dataset
        self.name = name
        self.commit_id = commit_id
        if self.commit_id == ROOT_COMMIT_ID:
            return

        self.parent_commit_id = parent_commit_id
        self.title = title
        self.description = description
        self.committer = committer
        self.committed_at = committed_at

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    def to_pyobj(self) -> Dict[str, str]:
        """Dump the instance to a python dict.

        Returns:
            A python dict containing all the information of the named commit::

                {
                    "name": <str>
                    "commit_id": <str>
                    "parent_commit_id": <str>
                    "title": <str>
                    "description": <str>
                    "committer":  <str>
                    "committed_at": <str>
                }

        """
        return self._dumps()


class CommitManager:
    """This class defines the operations on the commit in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(
        self, revision: Optional[str], offset: int, limit: int
    ) -> Generator[Commit, None, int]:
        if revision is None:
            revision = self._dataset.HEAD.commit_id

        response = list_commits(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            revision=revision,
            offset=offset,
            limit=limit,
        )

        for item in response["commits"]:
            yield Commit(self._dataset, **item)

        return response["total_count"]  # type: ignore[no-any-return]

    def get(self, revision: Optional[str] = None) -> Commit:
        """Get the certain commit with the given revision.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name. If it is not given, get the current commit.

        Returns:
            The :class:`.Commit` instance with the given revision.

        """
        if revision is None:
            revision = self._dataset.HEAD.commit_id

        response = get_revision(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            revision=revision,
        )
        return Commit(self._dataset, **response)

    def list(self, revision: Optional[str] = None) -> LazyPagingList[Commit]:
        """List the commits.

        Arguments:
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If it is given, list the commits before the given commit.
                If it is not given, list the commits before the current commit.

        Returns:
            The LazyPagingList of :class:`commits<.Commit>` instances.

        """
        return LazyPagingList(
            lambda offset, limit: self._generate(revision, offset, limit),
            24,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Tag and TagManager."""

from typing import TYPE_CHECKING, Generator, Optional

from graviti.exception import ResourceNotExistError
from graviti.manager.commit import NamedCommit
from graviti.manager.lazy import LazyPagingList
from graviti.openapi import create_tag, delete_tag, get_tag, list_tags

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Tag(NamedCommit):
    """This class defines the structure of the tag of a commit.

    Arguments:
        dataset: Class :class:`~graviti.dataset.dataset.Dataset` instance.
        name: The name of the tag.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.
        committed_at: The time when the draft is committed.

    """


class TagManager:
    """This class defines the operations on the tag in Graviti.

    Arguments:
        dataset: :class:`~graviti.manager.dataset.Dataset` instance.

    """

    def __init__(self, dataset: "Dataset") -> None:
        self._dataset = dataset

    def _generate(self, offset: int, limit: int) -> Generator[Tag, None, int]:
        response = list_tags(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            offset=offset,
            limit=limit,
        )

        for item in response["tags"]:
            yield Tag(self._dataset, **item)

        return response["total_count"]  # type: ignore[no-any-return]

    def create(self, name: str, revision: Optional[str] = None) -> Tag:
        """Create a tag for a commit.

        Arguments:
            name: The tag name to be created for the specific commit.
            revision: The information to locate the specific commit, which can be the commit id,
                the branch name, or the tag name.
                If the revision is not given, create the tag for the current commit.

        Returns:
            The :class:`.Tag` instance with the given name.

        """
        if not revision:
            revision = self._dataset.HEAD.commit_id

        response = create_tag(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            name=name,
            revision=revision,
        )
        return Tag(self._dataset, **response)

    def get(self, name: str) -> Tag:
        """Get the certain tag with the given name.

        Arguments:
            name: The required tag name.

        Raises:
            ResourceNotExistError: When the name is an empty string.

        Returns:
            The :class:`.Tag` instance with the given name.

        """
        if not name:
            raise ResourceNotExistError(resource="tag", identification=name)

        response = get_tag(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            tag=name,
        )
        return Tag(self._dataset, **response)

    def list(self) -> LazyPagingList[Tag]:
        """List the information of tags.

        Returns:
            The LazyPagingList of :class:`tags<.Tag>` instances.

        """
        return LazyPagingList(self._generate, 24)

    def delete(self, name: str) -> None:
        """Delete a tag.

        Arguments:
            name: The tag name to be deleted for the specific commit.

        Raises:
            ResourceNotExistError: When the name is an empty string.

        """
        if not name:
            raise ResourceNotExistError(resource="tag", identification=name)

        delete_tag(
            self._dataset.access_key,
            self._dataset.url,
            self._dataset.owner,
            self._dataset.name,
            tag=name,
        )



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Interfaces about the search."""

from typing import Any, Dict, Optional

from graviti.openapi.requests import open_api_do


def create_search(
    access_key: str,
    url: str,
    owner: str,
    dataset: str,
    *,
    commit_id: str,
    sheet: str,
    criteria: Dict[str, Any],
    offset: Optional[int] = None,
    limit: Optional[int] = None,
) -> Dict[str, Any]:
    """Execute the OpenAPI `POST /v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets\
    /{sheet}/search?offset={offset}&limit={limit}`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti website.
        owner: The owner of the dataset.
        dataset: Name of the dataset, unique for a user.
        commit_id: The commit id.
        sheet: The sheet name.
        criteria: The criteria of the search.
        offset: The offset of the page. The default value of this param in OpenAPIv2 is 0.
        limit: The limit of the page. The default value of this param in OpenAPIv2 is 128.

    Returns:
        The response of OpenAPI.

    Examples:
        >>> create_search(
        ...     "ACCESSKEY-********",
        ...     "https://gas.graviti.com/gatewayv2/tensorbay-open-api",
        ...     "czhual",
        ...     "BDD100K",
        ...     commit_id = "fde63f357daf46088639e9f57fd81cad",
        ...     sheet = "train",
        ...     criteria = {
        ...         "opt": "or",
        ...         "value": [
        ...             {
        ...                 "opt": "eq",
        ...                 "key": "filename",
        ...                 "value": "0000f77c-6257be58.jpg"
        ...             },
        ...             {
        ...                 "opt": "and",
        ...                 "value": [
        ...                     {
        ...                         "opt": "eq",
        ...                         "key": "attribute.weather",
        ...                         "value": "clear"
        ...                     },
        ...                     {
        ...                         "opt": "eq",
        ...                         "key": "attribute.timeofday",
        ...                         "value": "daytime"
        ...                     }
        ...                 ]
        ...             }
        ...         ]
        ...     }
        ... )
        {
            "data": [
                {
                    "__record_key": "123750493121329585",
                    "filename": "0000f77c-6257be58.jpg",
                    "image": {
                        "url": "https://content-store-prod-vers",
                        "checksum": "dcc197970e607f7576d978972f6fb312911ce005"
                    },
                    "attribute": {
                        "weather": "clear",
                        "scene": "city street",
                        "timeofday": "daytime"
                    }
                },
                ...(total 128 items)
            ],
            "offset": 0,
            "record_size": 128,
            "total_count": 200
        }

    """
    url = f"{url}/v2/datasets/{owner}/{dataset}/commits/{commit_id}/sheets/{sheet}/search"

    post_data = {"criteria": criteria}
    params = {}
    if offset is not None:
        params["offset"] = offset
    if limit is not None:
        params["limit"] = limit

    return open_api_do(  # type: ignore[no-any-return]
        "POST", access_key, url, json=post_data, params=params
    ).json()



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""Portex record field releated classes."""


from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, TypeVar, Union

import pyarrow as pa

from graviti.portex.base import _INDENT, PortexType
from graviti.portex.package import Imports
from graviti.utility import NameOrderedDict

_T = TypeVar("_T", bound="Fields")


class Fields(NameOrderedDict[PortexType]):
    """Represents a Portex ``record`` fields dict."""

    def __init__(self, fields: Union[Iterable[Tuple[str, PortexType]], Mapping[str, PortexType]]):
        super().__init__(fields)

    def __repr__(self) -> str:
        return self._repr1(0)

    @classmethod
    def _check_value(cls, value: PortexType) -> None:
        if not isinstance(value, PortexType):
            raise TypeError(f'The value in "{cls.__name__}" should be a PortexType')

    def _repr1(self, level: int) -> str:
        indent = level * _INDENT
        lines = ["{"]
        for name, portex_type in self.items():
            lines.append(
                f"{_INDENT}'{name}': "  # pylint: disable=protected-access
                f"{portex_type._repr1(level + 1)},"
            )

        lines.append("}")
        return f"\n{indent}".join(lines)

    def _setitem(self, key: str, value: PortexType) -> None:
        self._check_value(value)
        super()._setitem(key, value)

    @property
    def imports(self) -> Imports:
        """Get the Fields imports.

        Returns:
            The :class:`Imports` instance of this Fields.

        """
        imports = Imports()
        for portex_type in self._data.values():
            imports.update_from_type(portex_type)

        return imports

    def insert(self, index: int, name: str, portex_type: PortexType) -> None:
        """Insert the name and portex_type at the index.

        Arguments:
            index: The index to insert the field.
            name: The name of the field to be inserted.
            portex_type: The portex_type of the field to be inserted.

        Raises:
            KeyError: When the name already exists in the Fields.

        """
        self._check_key(name)
        self._check_value(portex_type)

        if self.__contains__(name):
            raise KeyError(f'"{name}" already exists in the Fields')

        self._keys.insert(index, name)
        self._data.__setitem__(name, portex_type)

    def astype(self, name: str, portex_type: PortexType) -> None:
        """Convert the type of the field with the given name to the new PortexType.

        Arguments:
            name: The name of the field to convert.
            portex_type: The new PortexType of the field to convert to.

        Raises:
            KeyError: When the name does not exist in the Fields.

        """
        self._check_value(portex_type)

        if not self.__contains__(name):
            raise KeyError(f'"{name}" does not exist in the Fields')

        self._data.__setitem__(name, portex_type)

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename the name of a field.

        Arguments:
            old_name: The current name of the field to be renamed.
            new_name: The new name of the field to assign.

        """
        self._check_key(new_name)

        self._data.__setitem__(new_name, self._data.pop(old_name))
        self._keys.__setitem__(self._keys.index(old_name), new_name)

    @classmethod
    def from_pyobj(
        cls, content: List[Dict[str, Any]], imports: Optional[Imports] = None
    ) -> "Fields":
        """Create Portex fields dict instance from python list.

        Arguments:
            content: A python list representing a Portex fields dict.
            imports: The imports of the Portex fields dict.

        Returns:
            A Portex fields dict instance created from the input python list.

        """
        return Fields((item["name"], PortexType.from_pyobj(item, imports)) for item in content)

    def to_pyobj(self) -> List[Dict[str, Any]]:
        """Dump the instance to a python list.

        Returns:
            A Python List representation of the fields dict.

        """
        return [{"name": name, **portex_type.to_pyobj(False)} for name, portex_type in self.items()]

    def to_pyarrow(self) -> List[pa.Field]:
        """Convert the fields to a list of PyArrow Field.

        Returns:
            A list of PyArrow Field representing the fields of Portex record.

        """
        return [pa.field(name, portex_type.to_pyarrow()) for name, portex_type in self.items()]

    def copy(self: _T) -> _T:
        """Get a copy of the fields.

        Returns:
            A copy of the fields.

        """
        return self.__class__((name, portex_type.copy()) for name, portex_type in self.items())



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the Sheets."""

from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Iterator, MutableMapping

from graviti.dataframe import DataFrame
from graviti.portex import PortexType
from graviti.utility import LazyFactory, ReprMixin

if TYPE_CHECKING:
    from graviti.manager.dataset import Dataset


class Sheets(MutableMapping[str, DataFrame], ReprMixin):
    """The basic structure of the Graviti sheets."""

    _data: Dict[str, DataFrame]
    _dataset: "Dataset"

    def __len__(self) -> int:
        return self._get_data().__len__()

    def __getitem__(self, key: str) -> DataFrame:
        return self._get_data().__getitem__(key)

    def __setitem__(self, key: str, value: DataFrame) -> None:
        self._get_data().__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._get_data().__delitem__(key)

    def __iter__(self) -> Iterator[str]:
        return self._get_data().__iter__()

    def _list_data(self, offset: int, limit: int, sheet_name: str) -> Dict[str, Any]:
        raise NotImplementedError

    def _list_sheets(self) -> Dict[str, Any]:
        raise NotImplementedError

    def _get_sheet(self, sheet_name: str) -> Dict[str, Any]:
        raise NotImplementedError

    def _init_data(self) -> None:
        self._data = {}
        sheets_info = self._list_sheets()["sheets"]

        for sheet_info in sheets_info:
            sheet_name = sheet_info["name"]
            sheet = self._get_sheet(sheet_name)
            schema = PortexType.from_yaml(sheet["schema"])

            factory = LazyFactory(
                sheet["record_count"],
                128,
                partial(self._list_data, sheet_name=sheet_name),
                schema.to_pyarrow(),
            )
            paging_lists = factory.create_lists(schema.get_keys())
            self._data[sheet_name] = DataFrame._from_paging(  # pylint: disable=protected-access
                paging_lists, schema
            )

    def _get_data(self) -> Dict[str, DataFrame]:
        if not hasattr(self, "_data"):
            self._init_data()

        return self._data



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The base elements of Portex type."""


import json
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple, Type, TypeVar

import pyarrow as pa
import yaml

from graviti.portex.package import Imports, Package
from graviti.portex.register import PyArrowConversionRegister
from graviti.utility import PathLike

if TYPE_CHECKING:
    from graviti.dataframe import Container
    from graviti.portex.param import Params

_INDENT = " " * 2

PYARROW_TYPE_ID_TO_PORTEX_TYPE = PyArrowConversionRegister.PYARROW_TYPE_ID_TO_PORTEX_TYPE

_T = TypeVar("_T", bound="PortexType")


class PortexType:
    """The base class of portex type."""

    nullable: bool
    package: ClassVar[Package[Any]]
    params: ClassVar["Params"]
    container: Type["Container"]

    def __repr__(self) -> str:
        return self._repr1(0)

    @classmethod
    def _from_pyarrow(cls, pyarrow_type: pa.DataType) -> "PortexType":
        raise NotImplementedError

    def _repr1(self, level: int) -> str:
        with_params = False
        indent = level * _INDENT
        lines = [f"{self.__class__.__name__}("]
        for name, parameter in self.params.items():
            attr = getattr(self, name)
            if attr != parameter.default:
                with_params = True
                lines.append(
                    f"{_INDENT}{name}="  # pylint: disable=protected-access
                    f"{attr._repr1(level + 1) if hasattr(attr, '_repr1') else repr(attr)},"
                )

        if with_params:
            lines.append(")")
            return f"\n{indent}".join(lines)

        return f"{lines[0]})"

    def get_keys(  # pylint: disable=no-self-use
        self, _: Optional[str] = None
    ) -> List[Tuple[str, ...]]:
        """Get the keys to locate the data.

        Returns:
            The keys to locate the data.

        """
        return []

    @property
    def imports(self) -> Imports:
        """Get the PortexType imports.

        Returns:
            The :class:`Imports` instance of this PortexType.

        """
        imports = Imports()

        for name in self.params:
            argument = getattr(self, name)
            if hasattr(argument, "imports"):
                imports.update(argument.imports)

            if isinstance(argument, PortexType):
                class_ = argument.__class__
                imports[class_.__name__] = class_

        return imports

    @classmethod
    def from_pyobj(
        cls: Type[_T], content: Dict[str, Any], _imports: Optional["Imports"] = None
    ) -> _T:
        """Create Portex type instance from python dict.

        Arguments:
            content: A python dict representing a Portex type.

        Returns:
            A Portex type instance created from the input python dict.

        """
        if _imports is None:
            _imports = Imports.from_pyobj(content.get("imports", []))

        class_: Type[_T] = _imports[content["type"]]  # type: ignore[assignment]

        assert issubclass(class_, cls)
        kwargs = {}
        for name, param in class_.params.items():
            kwarg = content.get(name, ...)
            if kwarg is not ...:
                kwargs[name] = param.load(kwarg, _imports)

        type_ = class_(**kwargs)  # type: ignore[call-arg]
        return type_

    @classmethod
    def from_pyarrow(cls, pyarrow_type: pa.DataType) -> "PortexType":
        """Create Portex type instance from PyArrow type.

        Arguments:
            pyarrow_type: The PyArrow type.

        Raises:
            TypeError: When the PyArrow type is not supported.

        Returns:
            The created Portex type instance.

        """
        pyarrow_type_id = (
            pyarrow_type.value_type.id
            if pyarrow_type.id == pa.lib.Type_DICTIONARY  # pylint: disable=c-extension-no-member
            else pyarrow_type.id
        )

        try:
            portex_type = PYARROW_TYPE_ID_TO_PORTEX_TYPE[pyarrow_type_id]
        except KeyError:
            raise TypeError(f'Not supported PyArrow type "{pyarrow_type}"') from None

        return portex_type._from_pyarrow(pyarrow_type)  # pylint: disable=protected-access

    @classmethod
    def from_json(cls, content: str) -> "PortexType":
        """Create Portex type instance from JSON string.

        Arguments:
            content: A JSON string representing a Portex type.

        Returns:
            A Portex type instance created from the input JSON string.

        """
        return cls.from_pyobj(json.loads(content))

    @classmethod
    def from_yaml(cls, content: str) -> "PortexType":
        """Create Portex type instance from YAML string.

        Arguments:
            content: A YAML string representing a Portex type.

        Returns:
            A Portex type instance created from the input YAML string.

        """
        return cls.from_pyobj(yaml.load(content, yaml.Loader))

    def to_pyobj(self, _with_imports: bool = True) -> Dict[str, Any]:
        """Dump the instance to a python dict.

        Returns:
            A python dict representation of the Portex type.

        """
        pydict: Dict[str, Any] = {}
        if _with_imports:
            imports_pyobj = self.imports.to_pyobj()
            if imports_pyobj:
                pydict["imports"] = imports_pyobj

        pydict["type"] = self.__class__.__name__
        for name, parameter in self.params.items():
            attr = getattr(self, name)
            if attr != parameter.default:
                pydict[name] = parameter.dump(attr)

        return pydict

    def to_json(self) -> str:
        """Dump the instance to a JSON string.

        Returns:
            A JSON representation of the Portex type.

        """
        return json.dumps(self.to_pyobj())

    def to_yaml(self) -> str:
        """Dump the instance to a YAML string.

        Returns:
            A YAML representation of the Portex type.

        """
        return yaml.dump(self.to_pyobj(), sort_keys=False)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        Return:
            The corresponding builtin PyArrow DataType.

        """
        raise NotImplementedError

    def copy(self: _T) -> _T:
        """Get a copy of the portex type.

        Returns:
            A copy of the portex type.

        """
        return deepcopy(self)


def read_yaml(path: PathLike) -> PortexType:
    """Read a yaml file into Portex type.

    Arguments:
        path: The path of the yaml file.

    Returns:
        A Portex type instance created from the input yaml file.

    """
    with open(path, encoding="utf-8") as fp:
        return PortexType.from_pyobj(yaml.load(fp, yaml.Loader))


def read_json(path: PathLike) -> PortexType:
    """Read a json file into Portex type.

    Arguments:
        path: The path of the json file.

    Returns:
        A Portex type instance created from the input json file.

    """
    with open(path, encoding="utf-8") as fp:
        return PortexType.from_pyobj(json.load(fp))



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The PyArrow extension types and arrays to represent PortexType."""

import json
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Type, TypeVar, Union, overload

import pyarrow as pa

from graviti.utility.file import RemoteFile

if TYPE_CHECKING:
    # https://pylint.pycqa.org/en/latest/technical_reference/c_extensions.html
    from pyarrow.lib import DataType


class ExtensionBase(pa.ExtensionType):  # type: ignore[misc]
    """The base class for the PyArrow representation of PortexType types.

    Arguments:
        name: The PortexType name.
        storage_type: The corresponding PyArrow type.
        nullable: Whether the PortexType is nullable.
        kwargs: Other keyword arguments to init the PortexType.

    """

    _T = TypeVar("_T", bound="ExtensionBase")

    _PYARROW_PACKAGE_NAME: ClassVar[str]

    def __init__(
        self,
        name: str,
        storage_type: "DataType",
        *,
        nullable: bool = False,
        **kwargs: Any,
    ):
        self._name = name
        self._nullable = nullable
        self._kwargs = kwargs
        pa.ExtensionType.__init__(self, storage_type, f"{self._PYARROW_PACKAGE_NAME}.{name}")

    def __arrow_ext_serialize__(self) -> bytes:
        return json.dumps(self._to_pyobj()).encode("utf-8")

    @classmethod
    def __arrow_ext_deserialize__(cls: Type[_T], storage_type: "DataType", serialized: bytes) -> _T:
        kwargs = json.loads(serialized.decode("utf-8"))
        return cls(**kwargs, storage_type=storage_type)

    def _to_pyobj(self) -> Dict[str, Any]:
        return {"name": self._name, "nullable": self._nullable, **self._kwargs}


class BuiltinExtension(ExtensionBase):
    """This class defines the PyArrow representation of PortexBuiltinType type."""

    _PYARROW_PACKAGE_NAME = "builtin"


class ExternalExtension(ExtensionBase):
    """This class defines the PyArrow representation of PortexExternalType type.

    Arguments:
        name: The PortexExternalType name.
        storage_type: The corresponding PyArrow type.
        url: The repo url which contains the PortexExternalType.
        revision: The repo revision which contains the PortexExternalType.
        nullable: Whether the PortexExternalType is nullable.
        kwargs: Other keyword arguments to init the PortexExternalType.

    """

    _PYARROW_PACKAGE_NAME = "external"

    def __init__(
        self,
        name: str,
        storage_type: "DataType",
        url: str,
        revision: str,
        *,
        nullable: bool = False,
        **kwargs: Any,
    ):
        self._url = url
        self._revision = revision
        super().__init__(name, storage_type, nullable=nullable, **kwargs)

    def _to_pyobj(self) -> Dict[str, Any]:
        return {**super()._to_pyobj(), "url": self._url, "revision": self._revision}


class GravitiExtension(ExtensionBase):
    """This class defines the PyArrow representation of PortexGravitiType type."""

    _PYARROW_PACKAGE_NAME = "graviti"


class RemoteFileArray(pa.ExtensionArray):  # type: ignore[misc]
    """This class defines the PyArrow representation of RemoteFileArray."""

    @overload
    def __getitem__(self, index: int) -> RemoteFile:
        pass

    @overload
    def __getitem__(self, index: slice) -> "RemoteFileArray":
        pass

    def __getitem__(self, index: Union[int, slice]) -> Union[RemoteFile, "RemoteFileArray"]:
        if isinstance(index, slice):
            return super().__getitem__(index)  # type: ignore[no-any-return]

        item = super().__getitem__(index).as_py()
        return RemoteFile(item["url"], item["checksum"])


class RemoteFileType(GravitiExtension):
    """This class defines the PyArrow representation of RemoteFileType type."""

    def __init__(self) -> None:
        super().__init__("file", pa.struct({"url": pa.string(), "checksum": pa.string()}))

    def __arrow_ext_class__(self) -> Type[RemoteFileArray]:  # pylint: disable=no-self-use
        return RemoteFileArray
